// This file is part of Lx.
//
// Copyright 2003-2006, 2008, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.helpers;

import uk.me.nxg.lx.reader.LxReader;

import java.io.IOException;
import java.io.Reader;
import java.io.PushbackReader;
import java.io.InputStreamReader;
import java.net.URL;

// Import _all_ the SAX classes and interfaces
import org.xml.sax.*;

/**
 * A parser which can be a SAX1 or SAX2 parser, and can take its input
 * either from Lx files or XML files.
 *
 * <p>When <code>parse()</code> is called, the parser examines the first
 * non-space character to determine the type of the input file.
 * It acts as a SAX1 or SAX2
 * parser depending on whether, prior to the <code>parse()</code>
 * call, the <code>setDocumentHandler()</code> or
 * <code>setContentHandler()</code> method is called (or indeed any
 * other methods specific to only one of the two interfaces).  It is
 * an error to call both SAX1 and SAX2 methods of this class, and the
 * object will throw an {@link IllegalArgumentException} in this case.
 *
 * <p>At present, the MultiParser <em>cannot</em> be reused -- you
 * cannot parse two different (SAX1/2, XML/Lx) streams with the same
 * object.
 *
 * <p>The support for features and properties is necessarily simple.
 * Since we do not know which parser we're going to use until we call
 * <code>parse()</code>, we can't know which features and properties
 * are supported.  Therefore we accept all properties and features,
 * specifically including the two required SAX
 * properties, <code>http://xml.org/sax/features/namespaces</code>
 * and <code>http://xml.org/sax/features/namespace-prefixes</code>,
 * which must be supported by whatever parser we eventually use.
 * Any properties and features are passed directly on to the underlying parser.
 *
 * @author Norman Gray
 * @version 0.5.3, 2012 August 5
 */
public class MultiParser
        implements org.xml.sax.XMLReader, org.xml.sax.Parser {

    /**
     * Once we have determined which interface we are supporting --
     * SAX1 Parser or SAX2 XMLReader, set this to either 1 or 2.
     */
    private int saxLevel = 0;

    // SAX-Standard feature URIs
    private static String SAXNS
            = "http://xml.org/sax/features/namespaces";
    private static String SAXNSP
            = "http://xml.org/sax/features/namespace-prefixes";

    private static javax.xml.parsers.SAXParserFactory parserFactory;

    private Transformer.Verbosity verbosity = Transformer.Verbosity.QUIET;

    /* XMLReader instance variables */
    private XMLReader realXMLReader;
    private ContentHandler contentHandler;
    
    /* Parser instance variables */
    private Parser realParser;
    private DocumentHandler documentHandler;
    private java.util.Locale locale;
    
    /* Common instance variables */
    private EntityResolver entityResolver;
    private DTDHandler dtdHandler;
    private ErrorHandler errorHandler;

    // Map of properties and features
    private java.util.Map readerSettings;

    /* Non-SAX methods. */

    // The verbosity specification coult potentially be specified as a
    // Feature, but that's a SAX2-only method, also this three-way
    // verbosity would require more complication.
    public void setVerbosity(Transformer.Verbosity verbosity) {
        this.verbosity = verbosity;
    }

    /* SAX2 XMLReader interface */

    private boolean sax_namespaces = true;
    private boolean sax_namespace_prefixes = false;

    public boolean getFeature (String name)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        setSAXLevel(2);
        if (readerSettings != null) {
            Object v = readerSettings.get(name);
            if (v != null) {
                if (! (v instanceof Boolean))
                    throw new SAXNotRecognizedException
                            ("Feature " + name + " not a boolean!");
                return ((Boolean)v).booleanValue();
            }
        }

        // so it's not a pre-set one.  Try SAX features
        if (name.equals(SAXNS))
            return sax_namespaces;
        if (name.equals(SAXNSP))
            return sax_namespace_prefixes;

        // nothing matched
        throw new SAXNotRecognizedException("Unrecognised SAX feature " + name);
    }

    public void setFeature (String name, boolean value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        setSAXLevel(2);

        if (name.equals(SAXNS))
            sax_namespaces = value;
        else if (name.equals(SAXNSP))
            sax_namespace_prefixes = value;
        else if (name.startsWith("http://xml.org/sax"))
                throw new SAXNotRecognizedException
                        ("Unrecognised SAX feature " + name);
        else {
            // save this up for later
            if (readerSettings == null)
                readerSettings = new java.util.HashMap();
            readerSettings.put(name, Boolean.valueOf(value));
        }
    }

    public Object getProperty (String name)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        setSAXLevel(2);
        if (readerSettings != null) {
            Object v = readerSettings.get(name);
            if (v != null) 
                return v;
        }

        // nothing matched
        throw new SAXNotRecognizedException
                ("Unrecognised SAX feature " + name);
    }
    
    public void setProperty (String name, Object value)
            throws SAXNotRecognizedException, SAXNotSupportedException {
        setSAXLevel(2);
        // save this up for later
        if (readerSettings == null)
            readerSettings = new java.util.HashMap();
        readerSettings.put(name, value);
    }

    public void setContentHandler (ContentHandler handler) {
        setSAXLevel(2);
        contentHandler = handler;
    }
    
    public ContentHandler getContentHandler () {
        setSAXLevel(2);
        return contentHandler;
    }

    /* For set/getEntityResolver, set/getDTDHandler,
       set/getErrorHandler, and parse, see below */

    /* End of SAX2 XMLReader interface */


    /* SAX1 Parser interface */

    public void setLocale (java.util.Locale locale)
            throws SAXException {
        setSAXLevel(1);
        this.locale = locale;
    }
    
    public void setDocumentHandler (DocumentHandler handler) {
        setSAXLevel(1);
        documentHandler = handler;
    }
    
    /* For set/getEntityResolver, set/getDTDHandler,
       set/getErrorHandler, and parse, see below */

    /* End of SAX1 Parser interface */

    /* Methods common to both interfaces */

    public void setEntityResolver (EntityResolver resolver) {
        entityResolver = resolver;
    }
    
    public EntityResolver getEntityResolver () {
        return entityResolver;
    }
    
    public void setDTDHandler (DTDHandler handler) {
        dtdHandler = handler;
    }
    
    public DTDHandler getDTDHandler () {
        return dtdHandler;
    }
    
    public void setErrorHandler (ErrorHandler handler) {
        errorHandler = handler;
    }
    
    public ErrorHandler getErrorHandler () {
        return errorHandler;
    }

    public void parse (String systemId)
            throws IOException, SAXException {
	parse(new InputSource(systemId));
    }
    
    /**
     * Parses the given input source.  By the time this is called, we
     * should know whether we are masquerading as a SAX1 or SAX2
     * object, and we throw a SAXException if we cannot tell (which
     * means that neither <code>setContentHandler</code> nor
     * <code>setDocumentHandler</code> has been called).
     *
     * <p>We also need to determine whether we are being asked to
     * read an XML stream or an Lx stream, and we do this by looking at the
     * first non-space character in the <code>InputSource</code>, which can
     * only be one of '&lt;', ';' or '('.
     *
     * @param input the source from which we are to read the object
     * to be parsed
     * @throws IOException on any IO errors
     * @throws SAXException if it is passed on from the actual parse
     * method, or if neither <code>setContentHandler</code> nor
     * <code>setDocumenthandler</code> has been called so far
     */
    public void parse (InputSource input)
            throws IOException, SAXException {
        if (saxLevel == 0)
            throw new SAXException
                    ("MultiParser.parse called before either setContentHandler or setDocumentHandler");

        assert saxLevel == 1 || saxLevel == 2;

        if (saxLevel == 1 && documentHandler == null)
            throw new SAXException
                    ("Parser.parse called before setDocumentHandler");
        if (saxLevel == 2 && contentHandler == null)
            throw new SAXException
                    ("XMLReader.parse called before setContentHandler");

        // Find what type of file we're being asked to parse.  Obtain
        // from the InputSource the thing we're trying to parse, then
        // step through it until we find a character that lets us
        // decide.  A complication is that the documentation for
        // InputSource says ``An InputSource object belongs to the
        // application: the SAX parser shall never modify it in any
        // way (it may modify a copy if necessary).'', so we can't
        // _modify_ the CharacterStream on the given InputSource.
        // Instead, we create a new InputSource, and put in it, using
        // setCharacterStream, a (possibly new) PushbackReader, which
        // we've examined.
        PushbackReader pr = null;

        if (input.getCharacterStream() != null) {

            // easy
            pr = new PushbackReader(input.getCharacterStream());

        } else if (input.getByteStream() != null) {

            InputStreamReader isr = null;
            if (input.getEncoding() != null) {
                // Does InputStreamReader recognise this?
                try {
                    isr = new InputStreamReader(input.getByteStream(),
                                                input.getEncoding());
                } catch (java.io.UnsupportedEncodingException e) {
                    // no luck
                }
            }
            if (isr == null) {
                isr = new InputStreamReader(input.getByteStream());
            }
            assert isr != null;
            pr = new PushbackReader(isr);
            
        } else {

            String sysid = input.getSystemId();
            if (sysid == null)
                throw new SAXException
                        ("MultiParser: InputSource has neither CharacterStream, ByteStream nor SystemID");
            try {
                URL url = new URL(new URL("file:."), sysid);
                if (url.getProtocol().equals("file")) {
                    Reader fr = new java.io.FileReader(url.getFile());
                    if (fr == null)
                        throw new SAXException
                                ("Found file URL, but file not found ("
                                 + url.getFile() + ")");
                    pr = new PushbackReader(fr);
                } else {
                    pr = new PushbackReader
                            (new InputStreamReader(url.openStream()));
                }
            } catch (java.net.MalformedURLException e) {
                throw new SAXException("Malformed URL: " + e);
            }

        }

        int ch;

        // So we have a reader.  Find the first non-whitespace
        // character in it, save it in ch, and push it back onto the stream.
        assert pr != null;
        ch = pr.read();
        while (ch >= 0 && Character.isWhitespace((char)ch))
            ch = pr.read();
        if (ch < 0)
            throw new SAXException("No text at all in " + input.getSystemId());
        pr.unread(ch);

        // Now look at what we've got in ch
        boolean isXML;
        switch ((char)ch) {
          case '<':
            isXML = true;
            break;

          case ';':
          case '(':
            isXML = false;
            break;

          default:
            throw new SAXException
                    ("InputSource " + input.getSystemId()
                     + " is neither XML nor Lx "
                     + "(first non-whitespace character is " + (char)ch
                     + ")");
        }

        if (verbosity.compareTo(Transformer.Verbosity.QUIET) > 0)
            System.err.println("MultiParser: processing " + input + " as " + (isXML ? "XML" : "Lx"));

        // Create a new InputSource, which is a copy of the old one,
        // but with a CharacterStream which is the PushbackReader
        // we've just been working with.  
        InputSource myinput = new InputSource();
        myinput.setCharacterStream(pr);
        myinput.setSystemId(input.getSystemId());
        myinput.setPublicId(input.getPublicId());

        // Now find a SAX2 parser, based on the file type we've established
        XMLReader sax2parser;
        try {
            if (isXML) {
                if (parserFactory == null) {
                    parserFactory
                            = javax.xml.parsers.SAXParserFactory.newInstance();
                    parserFactory.setNamespaceAware(true);
                }
                sax2parser = parserFactory.newSAXParser().getXMLReader();
            } else {
                sax2parser = new LxReader();
            }
        } catch (javax.xml.parsers.ParserConfigurationException e) {
            throw new SAXException("Error creating parser: " + e);
        } catch (uk.me.nxg.lx.LxException e) {
            throw new SAXException("Error creating LxReader: " + e);
        }

        if (saxLevel == 2) {

            realXMLReader = sax2parser;

            if (entityResolver != null)
                realXMLReader.setEntityResolver(entityResolver);
            if (dtdHandler != null)
                realXMLReader.setDTDHandler(dtdHandler);
            if (errorHandler != null)
                realXMLReader.setErrorHandler(errorHandler);

            realXMLReader.setFeature(SAXNS, sax_namespaces);
            realXMLReader.setFeature(SAXNSP, sax_namespace_prefixes);

            if (readerSettings != null) {
                // There were there any saved properties or features
                for (java.util.Iterator i=readerSettings.entrySet().iterator();
                     i.hasNext(); ) {
                    java.util.Map.Entry e = (java.util.Map.Entry)i.next();
                    String k = (String)e.getKey();
                    Object v = e.getValue();
                    try {
                        if (v instanceof Boolean)
                            realXMLReader
                                    .setFeature(k, ((Boolean)v).booleanValue());
                        else
                            realXMLReader.setProperty(k, v);
                    } catch (SAXException ex) {
                        // do nothing -- nothing we can do this late
                    }   
                }
            }

            // The contentHandler must be non-null, since we
            // checked this above
            assert contentHandler != null;
            realXMLReader.setContentHandler(contentHandler);

            realXMLReader.parse(myinput);
            
        } else {

            realParser = new org.xml.sax.helpers.XMLReaderAdapter(sax2parser);

            if (entityResolver != null)
                realParser.setEntityResolver(entityResolver);
            if (dtdHandler != null)
                realParser.setDTDHandler(dtdHandler);
            if (errorHandler != null)
                realParser.setErrorHandler(errorHandler);

            // The documentHandler must be non-null, since we
            // checked this above
            assert documentHandler != null;
            realParser.setDocumentHandler(documentHandler);
                
            realParser.parse(myinput);
        }

        return;
    }

    /* Private helper methods */

    /**
     * Sets the SAX level.  This is called by each of the non-common
     * methods in this class, to indicate whether we are masquerading
     * as the level 1 or level 2 class.  We can't be both, so if
     * there's a conflict, we object.  Do so by throwing an
     * {@link IllegalArgumentException}, since this is not checked.
     * 
     * @param level the SAX level of the method which calls it
     */
    private void setSAXLevel(int level) {
        assert level == 1 || level == 2;
        if (saxLevel != 0 && saxLevel != level)
            throw new IllegalArgumentException
                    ("Can't call SAX" + level
                     + " methods after earlier SAX" + saxLevel
                     + " methods");
        saxLevel = level;
        assert saxLevel == 1 || saxLevel == 2;
    }
}

