// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.

package uk.me.nxg.lx.helpers;

import java.io.File;
import java.io.Writer;
import java.io.IOException;

import org.xml.sax.*;

/**
 * Dumps a SAX stream into an unambiguous format.
 */
public class SaxDumpWriter
        implements org.xml.sax.ContentHandler {
    
    java.io.PrintWriter writer;
    boolean showQNames = false; // show QNames in output?

    /**
     * Constructs a new writer which sends output to a File
     * @param file the file to receive the SAX dump
     */
    public SaxDumpWriter(File file)
            throws IOException {
        writer = new java.io.PrintWriter(new java.io.FileWriter(file));
    }

    /**
     * Constructs a new writer which sends output to a Writer
     * @param writer the Writer to receive the SAX dump
     */
    public SaxDumpWriter(Writer writer) {
        this.writer = new java.io.PrintWriter(writer);
    }

    /**
     * Constructs a new writer which sends output to an OutputStream
     * @param stream the OutputStream to receive the SAX dump
     */
    public SaxDumpWriter(java.io.OutputStream stream) {
        writer = new java.io.PrintWriter(stream);
    }

    /**
     * Switch on or off display of QNames in output.
     * @param value true if QNames are to be shown
     * @return the previous value
     */
    public boolean showQNames(boolean value) {
        boolean t = showQNames;
        showQNames = value;
        return t;
    }

    /**
     * Writes nothing to the output
     */
    public void setDocumentLocator(Locator locator) {
        // do nothing
    }

    /**
     * Writes <code>d+</code> on the output
     */
    public void startDocument ()
            throws SAXException {
        writer.println("d+");
    }

    /**
     * Writes <code>d-</code> on the output
     */
    public void endDocument()
            throws SAXException {
        writer.println("d-");
        writer.flush();
        writer.close();
    }

    /**
     * Writes <code>p+ prefix uri</code> on the output
     */
    public void startPrefixMapping (String prefix, String uri)
            throws SAXException {
        writer.print("p+ ");
        writer.print(prefix);
        writer.print(' ');
        writer.println(uri);
    }
    
    /**
     * Writes <code>p- prefix</code> on the output
     */
    public void endPrefixMapping (String prefix)
            throws SAXException {
        writer.print("p- ");
        writer.println(prefix);
    }
    
    /**
     * Writes an element-start indicator on the output.
     *
     * <p>Depending on
     * whether `showQNames' is true or false (depending on whether 
     * {@link #showQNames} has been called with a true or false
     * value), then this displays <code>E+
     * qName/</code><em>&lt;element-name&gt;</em> or <code>e+
     * </code><em>&lt;element-name&gt;</em>, where
     * <em>&lt;element-name&gt;</em> is either
     * <code>{uri}localname</code> or just <code>localname</code>,
     * depending on whether <code>uri</code> is empty or not.
     *
     * <p>This is followed by the attributes in alphabetical order of
     * the attribute name, with the attribute names including the
     * QNames and URIs as above.
     */
    public void startElement (String uri, String localName,
			      String qName, Attributes atts)
            throws SAXException {
        if (showQNames) {
            writer.print("E+ ");
            writer.print(qName);
            writer.print('/');
        } else {
            writer.print("e+ ");
        }
        if (uri.length() == 0)
            writer.print(localName);
        else {
            writer.print('{');
            writer.print(uri);
            writer.print('}');
            writer.print(localName);
        }

        java.util.Set a = new java.util.TreeSet();
        for (int i=0; i<atts.getLength(); i++) {
            StringBuffer sb = new StringBuffer();
            String auri = atts.getURI(i);
            if (showQNames)
                sb.append(atts.getQName(i)).append('/');
            if (auri.length() == 0)
                sb.append(atts.getLocalName(i));
            else {
                sb.append('{').append(auri).append('}');
                sb.append(atts.getLocalName(i));
            }
            sb.append('=');
            sb.append(atts.getValue(i));
            a.add(sb.toString());
        }
        for (java.util.Iterator i = a.iterator(); i.hasNext(); ) {
            writer.print(' ');
            writer.print((String)i.next());
        }
        writer.println();
    }
        
    /**
     * Writes <code>e- </code><em>&lt;element-name&gt;</em> on the
     * output, with the <em>&lt;element-name&gt;</em> as described in
     * {@link #startElement}.
     */
    public void endElement (String uri, String localName,
			    String qName)
            throws SAXException {
        if (showQNames) {
            writer.print("E- ");
            writer.print(qName);
            writer.print('/');
        } else {
            writer.print("e- ");
        }
        if (uri.length() == 0)
            writer.print(localName);
        else {
            writer.print('{');
            writer.print(uri);
            writer.print('}');
            writer.print(localName);
        }
        writer.println();
    }

    /**
     * Writes <code>c </code> on the output, followed by the
     * character string with <code>\n</code>, <code>\r</code> and
     * <code>\t</code> escaped.
     */
    public void characters (char ch[], int start, int length)
            throws SAXException {
        writer.print("c ");
        for (int i=start; i<start+length; i++) {
            switch (ch[i]) {
              case '\n':
                writer.write("\\n");
                break;
              case '\r':
                writer.write("\\r");
                break;
              case '\t':
                writer.write("\\t");
                break;
              default:
                writer.write(ch[i]);
                break;
            }
        }
        writer.println();
    }

    /**
     * Displays ignorable whitespace, unambiguously.
     *
     * <p>Writes <code>w </code> on the output, followed by the
     * whitespace with <code>\n</code>, <code>\r</code> and
     * <code>\t</code> escaped, and any other whitespace characters
     * indicated by <code>\</code><code>unnn</code> notation.
     */
    public void ignorableWhitespace (char ch[], int start, int length)
            throws SAXException {
        writer.print("w");
        for (int i=start; i<start+length; i++) {
            switch (ch[i]) {
              case '\n':
                writer.write(" \\n");
                break;
              case '\r':
                writer.write(" \\r");
                break;
              case '\t':
                writer.write(" \\t");
                break;
              default:
                writer.write(" \\u");
                writer.print((int)ch[i]);
                break;
            }
        }
        writer.println();
    }
    
    /**
     * Writes <code>i target data</code> on the output
     */
    public void processingInstruction (String target, String data)
            throws SAXException {
        writer.print("i ");
        writer.print(target);
        writer.print(' ');
        writer.println(data);
    }

    /**
     * Writes <code>s name</code> on the output
     */
    public void skippedEntity (String name)
            throws SAXException {
        writer.print("s ");
        writer.println(name);
    }
}
