// This file is part of Lx.
//
// Copyright 2003-2008, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.helpers;

import uk.me.nxg.lx.LxException;
import java.io.File;
import java.io.OutputStreamWriter;
import java.io.FileOutputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.transform.Source;
import javax.xml.transform.sax.SAXSource;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;


/**
 * Simple XSLT transformer methods using an Lx stylesheet.  These can be used
 * either as convenience methods wrapping the standard Transformer
 * boilerplate (though it obviously accepts Lx input also), or else as
 * a {@link #main main function}.
 *
 * <p>We read and write everything as UTF-8
 *
 * <p>The transformer sets the feature
 * <code>http://xml.org/sax/features/namespace-prefixes</code> to
 * false, and the feature
 * <code>http://xml.org/sax/features/namespaces</code> to true.
 */
public class Transformer {

    /**
     * Output format possibilities:
     * <ul>
     * <li>XML indicates that a Transformer should produce its output in XML</li>
     * <li>LX indicates that a Transformer should produce its output as Lx</li>
     * <li>SAX indicates that the transformer should display out which
     * unambiguously displays its SAX structure.  This
     * format is not documented here, but should be reasonably
     * stable.  This is primarly for debugging purposes.</li>
     * </ul>
     */
    public enum Outputs { INVALID, XML, LX, SAX }

    // verbosities: enable with '--verbose' and '--debug'
    public enum Verbosity { QUIET, VERBOSE, DEBUG}
    private static Verbosity verbosity = Verbosity.QUIET;

    public static Map readerSettings;

    /**
     * A main function which can act as a standalone XSLT transformer.
     *
     * <p>Usage is simple:
     * <pre>
     * % java -jar lx.jar &lt;options&gt; [param=value...] \
     *         input.{xml|lx} [stylesheet.lx [output]]
     * </pre>
     * If the <code>stylesheet.lx</code> argument is missing, the
     * transformer performs an identity transformation to stdout.  If
     * <code>output</code> is missing, the output goes to stdout.
     *
     * <p>Any <code>param=value</code> arguments on the command line
     * (which may be freely intermixed with the file arguments) are
     * used to set the initial values of any corresponding
     * <code>xsl:param</code> elements within the stylesheet.
     *
     * <p>If a `parameter' looks like a URI, however, by starting
     * <code>http://...</code>, then is regarded as a feature or property to
     * be set on the underlying parser: if the value is either
     * <code>true</code> or <code>false</code> it is a feature set on the
     * parser, otherwise it is a property.  See the
     * {@link uk.me.nxg.lx.reader.LxReader} class for a
     * description of the available properties.
     *
     * <p>If one of the input files is given as '-', then that file is read from standard input.
     *
     * <p>The following command line, using the default identity transformation,
     * is a useful way of creating a reasonably readable <code>.xslt</code>
     * version of an Lx script:</p>
     * <pre>
     * % java -jar lx.jar http://ns.nxg.me.uk/lx/generated-prefix=xsl --indent \
     *     myscript.lx &gt;myscript.xslt
     * </pre>
     * And the following is the precisely analogous way of converting an
     * existing <code>.xslt</code> script to an <code>.lx</code> script:
     * <pre>
     * % java -jar lx.jar --lx myscript.xslt &gt;myscript.lx
     * </pre>
     *
     * <p>Options:
     * <dl>
     * <dt><code>--xml</code>
     * <dd>Output is XML (the default)
     * <dt><code>--indent</code></dt>
     * <dd>If output is XML, this option switches on output indenting</dd>
     * <dt><code>--lx</code>
     * <dd>Output is in Lx format
     * <dt><code>--sax</code>
     * <dd>Output is the SAX stream in an unambiguous format.  See
     * {@link uk.me.nxg.lx.helpers.SaxDumpWriter}.
     * </dl>
     * If none of the output-format options is present, and the output
     * filename is present, then the output file name extension
     * indicates the output format.  Otherwise, the default is XML.
     */
    public static void main(String args[]) {

        File inputFile = null;
        File stylesheetFile = null;
        File outputFile = null;
        Outputs outputFormat = Outputs.INVALID;
        boolean indentXML = false; // true if output XML is to be indented

        Pattern parameterPattern
                = java.util.regex.Pattern.compile
                ("^([a-zA-Z_][a-zA-Z0-9_.-]*)=(.*)");
        Pattern propertyPattern
                = java.util.regex.Pattern.compile
                ("^(http://[^=]*)=(.*)");
        Map parameterMap = null;

        for (java.util.Iterator i = java.util.Arrays.asList(args).iterator();
             i.hasNext(); ) {
            String arg = (String)i.next();

            if (arg.startsWith("--")) {
                if (arg.equals("--xml")) {
                    outputFormat = Outputs.XML;
                } else if (arg.equals("--lx")) {
                    outputFormat = Outputs.LX;
                } else if (arg.equals("--sax")) {
                    outputFormat = Outputs.SAX;
                } else if (arg.equals("--indent")) {
                    indentXML = true;
                } else if (arg.equals("--quiet")) {
                    verbosity = Verbosity.QUIET;
                } else if (arg.equals("--verbose")) {
                    verbosity = Verbosity.VERBOSE;
                } else if (arg.equals("--debug")) {
                    verbosity = Verbosity.DEBUG;
                } else {
                    Usage();
                }
            } else {
                Matcher param = parameterPattern.matcher(arg);
                if (param.matches()) {
                    // this is a parameter setting
                    if (parameterMap == null)
                        parameterMap = new java.util.HashMap();
                    if (verbosity.compareTo(Verbosity.QUIET) > 0)
                        System.err.println("param:" + param.group(1)
                                           + "=" + param.group(2));
                    parameterMap.put(param.group(1), param.group(2));
                } else {
                    Matcher prop = propertyPattern.matcher(arg);
                    if (prop.matches()) {
                        // this is a property setting
                        if (readerSettings == null)
                            readerSettings = new java.util.HashMap();
                        readerSettings.put(prop.group(1), prop.group(2));
                    } else {
                        // this is one of the file arguments
                        if (inputFile == null) {
                            inputFile = new File(arg);
                        } else if (stylesheetFile == null) {
                            stylesheetFile = new File(arg);
                        } else if (outputFile == null) {
                            outputFile = new File(arg);
                        } else {
                            Usage();
                        }
                    }
                }
            }
        }
            
        if (inputFile == null)
            Usage();

        if (outputFormat == Outputs.INVALID) {
            outputFormat = Outputs.XML; // default
            if (outputFile != null) {
                String fn = outputFile.getName();
                if (fn.endsWith(".xml"))
                    outputFormat = Outputs.XML;
                else if (fn.endsWith(".lx"))
                    outputFormat = Outputs.LX;
                else if (fn.endsWith(".sax"))
                    outputFormat = Outputs.SAX;        
            }
        }

        if (verbosity.compareTo(Verbosity.QUIET) > 0) {
            System.err.println("Transform "
                               + (inputFile == null ? "<stdin>" : inputFile)
                               + " -> "
                               + (outputFile == null ? "<stdout>" : outputFile)
                               + " using transformation "
                               + (stylesheetFile == null ? "<stdin>" : stylesheetFile));
            System.err.println("  Output format: " + outputFormat + ", parameters " + parameterMap);
        }

        int exitstatus = doTransformation(inputFile,
                                          stylesheetFile,
                                          outputFile,
                                          outputFormat,
                                          parameterMap,
                                          indentXML);

	System.exit(exitstatus);
    }

    static int doTransformation(File inputFile,
                                File stylesheetFile,
                                File outputFile,
                                Outputs outputFormat,
                                Map parameterMap,
                                boolean indentOutput) {

        int exitstatus = 1;     // failure by default
        try {

            // create a new transformer using the stylesheetFile if available
            javax.xml.transform.Transformer trans;
            if (stylesheetFile == null)
                trans = createLxTransformer(null, null);
            else
                trans = createLxTransformer(fileReader(stylesheetFile), stylesheetFile.getCanonicalPath());

            trans.setErrorListener(new LxErrorListener());

            if (outputFormat == Outputs.XML && indentOutput)
                trans.setOutputProperty(javax.xml.transform.OutputKeys.INDENT, "yes");

            // We want document() and co to be resolved relative to
            // the inputFile, as well as the location of the stylesheet
            trans.setURIResolver(new RelativeResolver(inputFile));

            // did we set any parameters above?
            if (parameterMap != null) {
                for (java.util.Iterator i = parameterMap.entrySet().iterator();
                     i.hasNext(); ) {
                    java.util.Map.Entry e = (java.util.Map.Entry)i.next();
                    trans.setParameter((String)e.getKey(), e.getValue());
                }
            }

            // get to work
            transform(fileReader(inputFile),
                      trans,
                      (outputFile == null
                       ? null
                       : new OutputStreamWriter(new FileOutputStream(outputFile), "UTF-8")),
                      outputFormat,
                      inputFile.toURI());

            exitstatus = 0;     // success
        } catch (LxException e) {
            System.err.println("Can't create LxReader for stylesheet: " + e);
        } catch (org.xml.sax.SAXException e) {
            System.err.println("Can't set feature: " + e);
        } catch (java.io.FileNotFoundException e) {
            System.err.println("Can't find file " + e);
        } catch (java.io.UnsupportedEncodingException e) {
            System.err.println("Eh? UTF-8 isn't supported?!: " + e);
        } catch (java.io.IOException e) {
            System.err.println("IO Exception: " + e);
	} catch (javax.xml.transform.TransformerException e) {
            System.err.println("Transformer exception: "
                               + e.getMessageAndLocation());
            //e.printStackTrace(System.err);
	}

        return exitstatus;
    }

    /**
     * Transform an input file using a stylesheet.  This is a
     * convenience method which wraps
     * {@link #transform(Reader,Reader,Writer,Outputs,java.net.URI)}.
     *
     * @param input the file which is to be transformed
     * @param stylesheet the file which contains the
     * stylesheet.  This may be null, in which case the method
     * performs an identity transformation.
     * @param output the destination file.  This may be null, in which case
     * the output is sent to <code>System.out</code>.  If this is
     * non-null, the file extension (.xml, .lx or .sax) determines the
     * format of the output file; the default is XML.
     */
    public static void transform(File input,
                                 File stylesheet,
                                 File output)
            throws LxException,
            org.xml.sax.SAXException,
            javax.xml.transform.TransformerException,
            java.io.FileNotFoundException,
            java.io.IOException {

        if (input == null)
            throw new IllegalArgumentException("first argument may not be null");

        if (stylesheet == null)
            transform(input,
                      createLxTransformer(null, null),
                      output);
        else
            transform(input,
                      createLxTransformer(fileReader(stylesheet),
                                          stylesheet.getCanonicalPath()),
                      output);
    }

    /**
     * Transform an input using a given transformer.  See
     * {@link #transform(Reader, Reader, Writer,Outputs,java.net.URI)}.
     */
    public static void transform(File input,
                                 javax.xml.transform.Transformer trans,
                                 File output)
            throws LxException,
            org.xml.sax.SAXException,
            javax.xml.transform.TransformerException,
            java.io.FileNotFoundException,
            java.io.IOException {

        if (input == null)
            throw new IllegalArgumentException
                    ("first argument may not be null");

        Outputs outputFormat = Outputs.XML;
        if (output != null) {
            String fn = output.getName();
            if (fn.endsWith(".xml"))
                outputFormat = Outputs.XML;
            else if (fn.endsWith(".lx"))
                outputFormat = Outputs.LX;
            else if (fn.endsWith(".sax"))
                outputFormat = Outputs.SAX;
        }

        transform(fileReader(input),
                  trans,
                  (output == null
                   ? null
                   : new OutputStreamWriter(new FileOutputStream(output), "UTF-8")),
                  outputFormat,
                  input.toURI());
    }

    /**
     * Transform an input using a stylesheet.
     *
     * @param input the SAX input which is to be transformed; must not be null
     * @param stylesheet the SAX stream which represents the
     * stylesheet.  This may be null, in which case the method
     * performs an identity transformation.
     * @param output the destination.  This may be null, in which case
     * the output is sent to <code>System.out</code>.
     * @param outputFormat one of the constants
     * <code>Outputs.XML</code>, <code>Outputs.LX</code> or
     * <code>Outputs.SAX</code>.
     * @param inputSystemId the system-id of the
     * input source.  This is potentially useful, as any relative URI
     * references within the input stream will be resolved against
     * this.  It must not be null
     *
     * @throws LxException if there is a problem setting up the Lx reader
     * @throws IllegalArgumentException if there are illegal arguments
     * (surprise!)
     * @throws org.xml.sax.SAXException if there are problems setting
     * up the SAX streams
     * @throws javax.xml.transform.TransformerException on any
     * transformer problems
     */
    public static void transform(Reader input,
                                 Reader stylesheet,
                                 Writer output,
                                 Outputs outputFormat,
                                 java.net.URI inputSystemId) 
            throws LxException,
            org.xml.sax.SAXException,
            javax.xml.transform.TransformerException {

        if (input == null)
            throw new IllegalArgumentException("Invalid call: null input");
        if (inputSystemId == null)
            throw new IllegalArgumentException
                    ("Invalid call: null inputSystemId");

        transform(input,
                  createLxTransformer(stylesheet, inputSystemId.toString()),
                  output,
                  outputFormat,
                  inputSystemId);
    }

    /**
     * Transform an input using a given transformer.
     * See {@link #transform(Reader, Reader, Writer, Outputs, java.net.URI)}
     */
    public static void transform(Reader input,
                                 javax.xml.transform.Transformer trans,
                                 Writer output,
                                 Outputs outputFormat,
                                 java.net.URI inputSystemId) 
            throws LxException,
            org.xml.sax.SAXException,
            javax.xml.transform.TransformerException {

        if (input == null)
            throw new IllegalArgumentException("Invalid call: null input");
        try {
            if (output == null)
                output = new java.io.OutputStreamWriter(System.out, "UTF-8");
        } catch (java.io.UnsupportedEncodingException e) {
            assert false;       // this can't happen -- UTF-8 is always supported
        }
        if (inputSystemId == null)
            throw new IllegalArgumentException("Invalid call: null inputSystemId");

        // Create the transform result 
        javax.xml.transform.Result result = null;
        switch (outputFormat) {
          case XML:
            result = new javax.xml.transform.stream.StreamResult(output);
            break;

          case LX:
            result = new javax.xml.transform.sax.SAXResult
                    (new LxWriter(output));
            break;

          case SAX:
            result = new javax.xml.transform.sax.SAXResult
                    (new SaxDumpWriter(output));
            break;

          default:
            throw new IllegalArgumentException
                    ("Impossible outputFormat: " + outputFormat);
        }

        assert result != null;
        assert input != null;

        // Create a SAX source for the transformer
        XMLReader inputReader = new uk.me.nxg.lx.helpers.MultiParser();
        inputReader.setFeature("http://xml.org/sax/features/namespace-prefixes", false);
        inputReader.setFeature("http://xml.org/sax/features/namespaces", true);
        ((uk.me.nxg.lx.helpers.MultiParser)inputReader).setVerbosity(verbosity);
        if (readerSettings != null) {
            for (java.util.Iterator i=readerSettings.entrySet().iterator();
                 i.hasNext(); ) {
                java.util.Map.Entry e = (java.util.Map.Entry)i.next();
                String k = (String)e.getKey();
                String v = (String)e.getValue();
                if (v.equals("true"))
                    inputReader.setFeature(k, true);
                else if (v.equals("false"))
                    inputReader.setFeature(k, false);
                else
                    inputReader.setProperty(k, v);
            }
        }       

        InputSource saxInputSource = new InputSource(input);
        // Set the sysid of the input file so that relative
        // references to eg, DTDs, can be resolved by parsers and
        // transformers.
        assert inputSystemId != null;
        saxInputSource.setSystemId(inputSystemId.toString());

        Source inputSource = new SAXSource(inputReader, saxInputSource);

        // Do the transformation
        trans.transform(inputSource, result);
    }

    /**
     * Create a {@link javax.xml.transform.Transformer} from an Lx
     * stylesheet.
     *
     * @param stylesheet the Lx stylesheet which is to be made into a
     * Transformer; if this is null, then return a transformer which performs
     * the identity transformation
     * @param srcSysId the system ID of the stylesheet.  If
     * <code>stylesheet</code> is non-null, this must be non-null also.
     */
    public static javax.xml.transform.Transformer
            createLxTransformer(Reader stylesheet, String srcSysId)
            throws uk.me.nxg.lx.LxException,
            javax.xml.transform.TransformerException,
            org.xml.sax.SAXException {

        if (stylesheet != null && srcSysId == null)
            throw new IllegalArgumentException
                    ("Null source system ID for non-null stylesheet");
        
        // Create the reader for the stylesheet -- accept only .lx
        // files, rather than using MultiParser
        XMLReader stylesheetReader = new uk.me.nxg.lx.reader.LxReader();
        stylesheetReader
                .setFeature("http://xml.org/sax/features/namespace-prefixes",
                            false);
        stylesheetReader
                .setFeature("http://xml.org/sax/features/namespaces",
                            true);

        if (verbosity == Verbosity.DEBUG)
            ((uk.me.nxg.lx.reader.LxReader)stylesheetReader).setVerbose(true);

        javax.xml.transform.Transformer trans;
        if (stylesheet == null) {
            trans = javax.xml.transform.TransformerFactory
                    .newInstance().newTransformer();
        } else {
            InputSource inputsrc = new InputSource(stylesheet);
            inputsrc.setSystemId(srcSysId);
            Source tranSrc
                    = new SAXSource(stylesheetReader, inputsrc);
            trans = javax.xml.transform.TransformerFactory.newInstance().newTransformer(tranSrc);
        }
        return trans;
    }

    /**
     * Create a Reader from a File, making sure that we read using UTF-8.
     * If the File's name is "-", then read from System.in.
     */
    private static Reader fileReader(File f)
            throws java.io.FileNotFoundException {
        try {
            if (f.getName().equals("-")) { // special case
                return new java.io.BufferedReader(new java.io.InputStreamReader(System.in));
            } else {
                return new java.io.BufferedReader(new java.io.InputStreamReader(new java.io.FileInputStream(f), "UTF8"));
            }
        } catch (java.io.UnsupportedEncodingException e) {
            assert false;       // can't happen -- UTF-8 is always supported
            return null;
        }
    }

    private static void Usage() {
        System.err.println("Usage: java -jar lx.jar [options] [parameters] input [transform [output]]");
        System.err.println();
        System.err.println("Options:");
        System.err.println("  --xml    : produce output in XML syntax");
        System.err.println("  --lx     : produce output in Lx syntax");
        System.err.println("  --sax    : produce (debugging) output showing the SAX stream");
        System.err.println("  --indent : indent output");
        System.err.println();
        System.err.println("[parameters]");
        System.err.println("  x=y                   : set the stylesheet parameter 'x' to value 'y'");
        System.err.println("  http://...=foo        : set the given property to be 'foo'");
        System.err.println("  http://...=true/false : set the given feature");
        System.err.println();
        System.err.println("The input can be in XML or Lx syntax.");
        System.err.println("Input files can be given as '-' to read from standard input.");
        System.err.println("If the transform argument is omitted, it defaults to an identity transform.");

        System.exit(1);
    }

    /**
     * A class which provides an alternative URIResolver.
     */
    private static class RelativeResolver
            implements javax.xml.transform.URIResolver {
        java.net.URI baseURI;   // the base file as a URI

        /**
         * Create a resolver which will resolve relative URIs against
         * a given base.  It tries against this base file before
         * trying against the base file passed to the resolve()
         * method.
         *
         * <p>Hmm: would it be better if we did these in the other order?
         * @param baseFile the File which is to be used as the base
         * URI for subsequent entity resolutions
         */
        RelativeResolver (File baseFile) {
            baseURI = baseFile.getAbsoluteFile().toURI();
        }
        public Source resolve(String href, String base) {
            try {
                java.io.FileInputStream fis;
                java.net.URI hrefURI = new java.net.URI(href);
                java.net.URI abshrefURI = baseURI.resolve(hrefURI);
                try {
                    fis = new java.io.FileInputStream(new File(abshrefURI));
                } catch (java.io.FileNotFoundException e) {
                    // Didn't find that -- so try again with 'base' as
                    // the base URI
                    abshrefURI = (new java.net.URI(base)).resolve(hrefURI);
                    try {
                        fis = new java.io.FileInputStream(new File(abshrefURI));
                    } catch (java.io.FileNotFoundException e2) {
                        System.err.println("Can't find file <" + href
                                           + "> (looked near " + baseURI + " and " + base);
                        fis = null;
                    }
                }
                if (fis == null)
                    return null;
                else
                    return new SAXSource(new InputSource(new java.io.InputStreamReader(fis, "UTF8")));
            } catch (java.net.URISyntaxException e) {
                // this really shouldn't happen
                System.err.println("RelativeResolver: malformed URI: " + e);
                return null;
            } catch (java.io.UnsupportedEncodingException e) {
                assert false;   // can't happen -- UTF8 is always supported
                return null;
            }
        }
    }

    private static class LxErrorListener implements javax.xml.transform.ErrorListener {
        public void error(javax.xml.transform.TransformerException e) {
            System.err.println("Error during Lx transform: " + e);
            e.printStackTrace();
        }
        public void fatalError(javax.xml.transform.TransformerException e) {
            System.err.println("Fatal error during Lx transform: " + e);
        }
        public void warning(javax.xml.transform.TransformerException e) {
            System.err.println("Warning during Lx transform: " + e);
        }
    }
}
