// This file is part of Lx.
//
// Copyright 2003, 2004, 2005, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

import java.util.List;
import java.util.Iterator;


/**
 * Represents an element.
 *
 * <p>The localname, prefix, QName and URI
 * returned by the respective methods all return Strings which have
 * been interned with {@link java.lang.String#intern}.
 */
class Element {
    private QName elementQName;
    private Attributes atts = new Attributes();
    
    /**
     * Creates a new element, by reading the required information from
     * the given Tokenizer.  Leaves the Tokenizer ready to read the
     * next token after the list of attributes.
     *
     * <p>The prefix information in the QName we read is interpreted
     * by the {@link QName} object we create -- it is this that
     * handles the namespacing declarations that have built up on the stack.
     *
     * @param t a Tokenizer from which an element is to be read; the
     * next thing to be read from the Tokenizer should be the
     * open-parenthesis which indicates a new element
     * @param stack an element/namespace stack which provides context
     * for features and namespace mappings
     * @throws LxException if there is an error parsing the stream
     */
    public Element(Tokenizer t, Stack stack)
            throws LxException {
        Token nextToken = null;//Token.BAD_TOKEN;

        try {
            nextToken = t.getToken();
            if (nextToken != Token.OPENPAREN)
                throw new LxException("Element() expected to read '('", t);

            nextToken = t.getToken();
            if (nextToken != Token.SYMBOL)
                throw new LxException
                        ("Unexpected token type at start of Element: '"
                         + t.tokenToString(nextToken)
                         + "' (require symbol)", t);

            elementQName = new QName(t.value(), stack);

            if (stack.getFeature(LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE)
                && stack.getFeature(LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE))
                throw new LxException
                        ("reader in illegal state: features "
                         + LxReader.LX_ATTRIBUTES_INHERIT_NAMESPACE
                         + " and "
                         + LxReader.LX_ATTRIBUTES_IN_NO_NAMESPACE
                         + " are both true",
                         t);

            // Now read the attributes
            boolean foundKeywords = gobbleKeywordValuePairs(t, stack);

            nextToken = t.getToken();

            // Are there any (@ (...)) attributes?
            if (nextToken == Token.OPENPAREN) {
                // peek ahead, to see if the next thing is an '@' symbol
                Token nextNextToken = t.getToken();
                if (nextNextToken == Token.SYMBOL
                    && t.value().equals("@")) {
                    // bingo
                    if (foundKeywords) {
                        // not so fast, buster -- we already found
                        // some attributes
                        throw new LxException
                                ("Can't specify (@...) attributes after keyword attributes", t);
                    }

                    gobbleAttributePairs(t, stack);
                    nextToken = t.getToken(); // should finish the (@...) list
                    if (nextToken != Token.CLOSEPAREN) {
                        throw new LxException
                          ("Unexpected end of attribute list -- expected ')'",
                           t);
                    }
                } else {
                    // No, this isn't an attribute list -- back up two tokens
                    t.ungetToken(nextNextToken);
                    t.ungetToken(nextToken);
                }
            } else {
                //assert(nextToken != Token.BAD_TOKEN);
                t.ungetToken(nextToken);
            }

            // we don't need to check that there are no forthcoming
            // keyword attributes, because that will be caught by the
            // general prohibition of keywords in element content
            
        } catch (java.io.IOException e) {
            throw new LxException(e);
        }
    }

    /**
     * Obtains the local name of this Element's GI.
     *
     * @return the element's GI as an XML LocalName
     */
    public String localName() {
        assert elementQName != null;
        String t = elementQName.localName();
        assert t != null;
        return t;
    }

    /**
     * Obtains the QName of this Element's GI.
     *
     * @return the element's GI as an XML QName
     */
    public String qName() {
        assert elementQName != null;
	String t = elementQName.qName();
        assert t != null;
        return t;
    }

    /**
     * Obtains the namespace of this Element's GI.
     *
     * @return the element's namespace, or null if it is
     * in no namespace
     */
    public String namespace() {
        assert elementQName != null;
        String t = elementQName.namespace();
        return t == null ? "" : t;
    }

    /**
     * Get the attribute list for this element
     *
     * @return the map of attributes
     */
    public Attributes atts() {
	return atts;
    }

    /**
     * Find the next attribute which is unspecified but required.
     * @param attinfo information about the required attributes,
     * which may be null, in which case the method always returns null
     * @return the next required attribute, or null if there are no
     * unspecified attributes left
     */
    private QName nextRequiredAttribute(Stack.AttributeInfo attinfo) {
        if (attinfo == null)
            return null;
        for (Iterator attiter = attinfo.allRequired(); attiter.hasNext(); ) {
            QName reqatt = (QName)attiter.next();
            if (! atts.isPresent(reqatt))
                return reqatt;
        }
        return null;
    }

    /**
     * Gobble a list of keyword-value pairs.  Leaves the tokenizer
     * stream pointing after any which are consumed.
     *
     * @param t the Tokenizer representing the input stream
     * @return true if any keyword-value pairs were encountered
     */
    private boolean gobbleKeywordValuePairs(Tokenizer t, Stack stack)
            throws LxException, java.io.IOException {
        // Handle any DTD declarations
        Stack.AttributeInfo attinfo = stack.getElementDecl(elementQName);

        boolean foundKeywordValuePairs = false;
        Token nextToken = null;//Token.BAD_TOKEN;

        boolean readAllKeywordAttributes = false;
        while (!readAllKeywordAttributes) {

            nextToken = t.getToken();
            if (nextToken == Token.KEYWORD) {
                String kw = t.value();
                if (kw.startsWith("xml")) {
                    // (includes "xmlns": reserved to XML, and not
                    // used for namespacing here
                    throw new LxException
                            ("Found (reserved) xml attribute in Lx input",
                             t);
                }
                foundKeywordValuePairs = true;

                Token valueToken = t.getToken();

                if (valueToken == Token.STRING
                    || valueToken == Token.SYMBOL) {
                    // that's OK
                    QName attQN = new QName(kw, stack, elementQName);
                    if (attinfo != null
                        && !attinfo.allowedAttribute(attQN)) {
                        // There are some declared attributes for
                        // this element: this isn't one of them
                        throw new LxException
                                ("Element " + elementQName
                                 + " has declared attributes, but "
                                 + attQN + " not recognised", t);
                    }

                    // have we been here before?
                    if (atts.isPresent(attQN))
                        throw new LxException
                                ("Element " + elementQName
                                 + ": can't add duplicate attribute "
                                 + attQN, t);
                    // either this element has no declared attributes,
                    // or this attribute is OK
                    atts.add(attQN, t.value());

                } else {
                    throw new LxException
                            ("Unexpected type for keyword value: '"
                             + t.tokenToString(valueToken)
                             + "' (require string or symbol)", 
                             t);
                }

            } else if (nextToken == Token.STRING
                       || nextToken == Token.SYMBOL) {

                QName reqatt = nextRequiredAttribute(attinfo);
                if (reqatt == null)
                    readAllKeywordAttributes = true;
                else
                    atts.add(reqatt, t.value());

            } else if (nextToken == Token.CLOSEPAREN) {

                if (nextRequiredAttribute(attinfo) == null) 
                    readAllKeywordAttributes = true;
                else
                    throw new LxException
                            ("Unexpected end of attribute list", t);
                    
            } else {

                readAllKeywordAttributes = true;
                    
            }
        }
        t.ungetToken(nextToken);

        return foundKeywordValuePairs;
    }
    
    /**
     * Reads the contents of an (@...) attribute list, which should be a
     * list of zero or more pairs.  The tokenizer should start off
     * being positioned just before the list of pairs, and is left
     * after the list, but before the ')' which terminates the (@...)
     * construct.
     *
     * @param t the Tokenizer representing the input stream
     * @return true if a non-null set of attribute pairs was in fact
     * encountered (not actually very useful information)
     */
    private boolean gobbleAttributePairs(Tokenizer t, Stack stack)
            throws LxException, java.io.IOException {

        boolean foundAttributePairs = false;

        // Handle any DTD declarations
        Stack.AttributeInfo attinfo = stack.getElementDecl(elementQName);

        boolean readAllPairs = false;
        while (!readAllPairs) {
            Token token = t.getToken();
            if (token == Token.OPENPAREN) {
                // about to read a pair
                token = t.getToken();
                if (token == Token.SYMBOL) {
                    // that's OK
                    foundAttributePairs = true;
                    String attname = t.value();
                    if (attname.startsWith("xml")) {
                        // (includes "xmlns": reserved to XML, and not
                        // used for namespacing here
                        throw new LxException
                                ("Found (reserved) xml attribute in Lx input",
                                 t);
                    }

                    Token valueToken = t.getToken();
                    if (valueToken == Token.STRING
                        || valueToken == Token.SYMBOL) {
                        // that's OK
                        QName attQN = new QName(attname, stack, elementQName);
                        if (attinfo != null
                            && !attinfo.allowedAttribute(attQN)) {
                            // There are some declared attributes for
                            // this element: this isn't one of them
                            throw new LxException
                                    ("Element " + elementQName
                                     + " has declared attributes, but "
                                     + attQN + " not recognised", t);
                        }

                        // have we been here before?
                        if (atts.isPresent(attQN))
                            throw new LxException
                                    ("Element " + elementQName
                                     + ": can't add duplicate attribute "
                                     + attQN, t);

                        // either this element has no declared attributes,
                        // or this attribute is OK
                        atts.add(attQN, t.value());

                        // read the ')' which terminates the pair list
                        token = t.getToken();
                        if (token != Token.CLOSEPAREN) {
                            throw new LxException
                                    ("Malformed (@...) attribute pair -- too many tokens",
                                     t);
                        }
                    } else {
                        throw new LxException
                                ("Unexpected type for attribute value: '"
                                 + t.tokenToString(valueToken)
                                 + "' (require string or symbol)", 
                                 t);
                    }
                } else {
                    throw new LxException
                            ("Malformed (@...) attribute list -- first in pair should be symbol, not "
                             + t.tokenToString(token),
                             t);
                }
            } else if (token == Token.CLOSEPAREN) {
                // end of list -- put this ')' back on the stream and
                // exit the loop
                readAllPairs = true;
                t.ungetToken(token);
            } else {
                // malformed
                throw new LxException
                        ("Malformed (@...) attribute list -- should contain pairs, found " + t.tokenToString(token),
                         t);
            }
        } // end of loop
        
        return foundAttributePairs;
    }

}
