// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;


/**
 * Implements a lisp-like list.
 */
class LxList extends LxObject {

    // Though the following is implemented as a pair, we (currently)
    // assert that the cdr must always be a list, so that non-list
    // pairs are (again, currently) impossible.

    // car and cdr must end up both non-null
    private LxObject car;
    private LxList cdr;

    private static int listcounter = 0;

    public static LxList EMPTYLIST;
    static {
        class EmptyLxList extends LxList {
            EmptyLxList() {
                // empty
            }
            boolean isList() {
                return true;
            }
            boolean isPair() {
                return false;
            }
            boolean isNull() {
                return true;
            }
            int length() {
                return 0;
            }
            LxObject car() {
                return null;
            }
            LxList cdr() {
                return null;
            }
            public String toString() {
                return "()";
            }
        }
        EMPTYLIST = new EmptyLxList();
    }

    /**
     * The no-argument constructor.  The EmptyLxList
     * constructor in the static block above implicitly uses it, but
     * the code below must not.
     *
     * XXX Is it possible to have this no-argument constructor throw
     * an assertion error, if the EmptyLxList constructor above can
     * avoid invoking it?  Check details of object construction.
     *
     * The object which is constructed by the default constructor
     * _fails_ the assertions below.
     */
    private LxList() {
        // Nothing
    }

    /**
     * Creates a new LxList.
     *
     * @param o an object which is to become the car of the new
     * list.  This must be non-null -- passing it as null triggers an
     * assertion error
     */
    private LxList(LxObject o) {
        car = o;
        cdr = EMPTYLIST;
        assert car != null && cdr != null && cdr.isList();
    }

    boolean isPair() {
        return true;
    }

    boolean isList() {
        // It's currently impossible to construct non-list pairs, so this
        // is trivial
        assert car != null && cdr != null && cdr.isList();
        return cdr.isList();
    }
    
    boolean isNull() {
        return false;
    }

    LxObject car() {
        assert car != null;
        return car;
    }

    LxList cdr() {
        assert cdr != null;
        return cdr;
    }

    int length() {
        assert car != null && cdr != null && cdr.isList();
        // null list has length 0
        return cdr.length()+1;
    }

//     /**
//      * Append the given object to the end of this list.  This is not
//      * a strictly lispish append, since if the object is an LxObject
//      * as opposed to a LxList, then what is appended is a new list
//      * with this LxObject in its car.  That is, we guarantee that the
//      * list remains a proper list.
//      *
//      * <p>Another non-lispishness is that this mutates the list.
//      * Note however that the overriding version of this on
//      * EmptyLxList returns a new LxList.
//      */
//     LxObject append(LxObject o) {
//         if (this == EMPTYLIST)
//             return o;
//         if (o == EMPTYLIST)
//             return this;             // nothing to do
//         if (cdr == EMPTYLIST) {
//             if (o instanceof LxList)
//                 cdr = o;
//             else {
//                 assert o instanceof LxObject;
//                 cdr = new LxList(o);
//             }
//             return this;
//         } else {
//             return cdr.append(o);
//         }
//     }
    
    public String toString() {
        StringBuffer sb = new StringBuffer("(");
        boolean first=true;
        for (LxList l=this; l != LxList.EMPTYLIST; l=l.cdr()) {
            if (first)
                first = false;
            else
                sb.append(' ');
            sb.append(l.car().toString());
        }
        sb.append(")");
        return sb.toString();
    }

    /**
     * Reads a list from the given {@link Tokenizer}.  The tokenizer
     * must be positioned just before the bracket opening the list.
     *
     * @param t a Tokenizer from which a list may be read
     * @return a list read from the token stream
     * @throws LxException if the list is malformed
     */
    static LxList readLxList(Tokenizer t)
            throws LxException {
        try {
            if (t.getToken() != Token.OPENPAREN)
                throw new LxException
                        ("readLxList: malformed list: expected '('");
        } catch (java.io.IOException e) {
            throw new LxException
                    ("readLxList: malformed list: " + e);
        }

        return getListContents(t);
    }

    /**
     * Reads the contents of a list, returning it as a list.  Reads
     * up to and including the ')' which terminates the list.
     *
     * @param t a Tokenizer positioned after the '(' which begins the list
     * @return a LxList containing the elements of the list
     * @throws LxException if the list is malformed
     */
    static private LxList getListContents(Tokenizer t)
            throws LxException {
        try {
            LxObject obj = LxObject.readLxObject(t);
            LxList l;
            if (obj == null) {
                // No object was available to be read.  Read the next
                // token: it should be a ')'.  If it's EOF that's
                // unexpected, since we're still in a list.  It can't
                // be anything else, since readLxObject would have
                // picked it up.
                listcounter=0;
                Token token = t.getToken();
                if (token != Token.CLOSEPAREN) {
                    // Ooops!
                    assert (token == Token.EOF);
                    throw new LxException("readLxList: Unexpected EOF");
                }
                // Fine!  This is an empty list
                l = EMPTYLIST;
            } else {
                l = new LxList(obj);
                l.cdr = getListContents(t);
            }

            assert l == EMPTYLIST
                    || (l.car != null && l.cdr != null && l.cdr.isList());
        
            return l;
        } catch (java.io.IOException e) {
            throw new LxException
                    ("readLxList: malformed: IOException reading list: " + e);
        }
    }
}
