// This file is part of Lx.
//
// Copyright 2003, 2004, 2008, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

/**
 * Implements a lisp-like set of objects.  Object, pair, list.
 */
class LxObject {

    private String value;
    private Token type;

    /**
     * Constructs a null object.  Private constructor -- generate new
     * instances with static method <code>LxObject.readLxObject</code>
     */
    protected LxObject() {
        type = null;//Token.BAD_TOKEN;
    }

    /**
     * Constructs a LxObject of a certain type.  Private constructor
     * -- generate new instances with {@link #readLxObject}
     */
    private LxObject(String value, Token type) {
        this.value = value;
        this.type = type;
    }

    boolean isSymbol() {
        return type == Token.SYMBOL;
    }

    boolean isKeyword() {
        return type == Token.KEYWORD;
    }

    boolean isString() {
        return type == Token.STRING;
    }

    boolean isToken() {
        return type == Token.TOKEN;
    }

    boolean isPair() {
        return false;
    }
    
    boolean isList() {
        return false;
    }

    /**
     * Returns true if this object is a string or a symbol
     */
    boolean isStringLike() {
        return type == Token.STRING || type == Token.SYMBOL;
    }

    /**
     * Return false if this is the token <code>#f</code>, true otherwise.
     */
    boolean isTrue() {
        return ! (type == Token.TOKEN && value.equals("f"));
    }

    public String toString() {
        return value == null ? "" : value;
    }

    /**
     * Reads an LxObject from the given {@link Tokenizer}.  The
     * Tokenizer must supply either a single object or a list (namely
     * a '(', then zero or more objects, then a ')').
     *
     * @param t a Tokenizer positioned before the object to be read
     * @return a LxObject read from the token stream.  If
     * there are no more objects to be read (that is, if the next
     * token is either a right-parenthesis or EOF), then return null;
     * in this case the Tokenizer position is unchanged.  Otherwise,
     * the Tokenizer is left positioned after the newly read LxObject.
     * @throws LxException if the tokens read from the stream are malformed
     */
    static LxObject readLxObject(Tokenizer t)
            throws LxException {
        try {
            LxObject ret;
            Token token;
        
            switch (token = t.getToken()) {
              case OPENPAREN:
                t.ungetToken(token);
                ret = (LxObject) LxList.readLxList(t);
                break;

              case CLOSEPAREN:
                t.ungetToken(token);
                ret = null;
                break;

              case EOF:
                ret = null;
                break;

              case SYMBOL:
              case TOKEN:
              case KEYWORD:
              case STRING:
                ret = new LxObject(t.value(), token);
                break;

              default:
                ret = null;         // to keep the compiler happy!
                assert(false);
            }

            return ret;
        } catch (java.io.IOException e) {
            throw new LxException("Error reading LxObject: " + e);
        }
    }
}
