// This file is part of Lx.
//
// Copyright 2003-2006, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;

import org.xml.sax.SAXNotRecognizedException;
import java.util.Map;
import java.util.Iterator;


/**
 * A stack.
 */
class Stack {

    private StackElement stacktop;
    private Mappings basemap;

    /** Counter for private method {@link #doPrefixMap} */
    private int prefixnum = -1;
    private String prefixbase = null;

    Stack() {
        basemap = new Mappings();
        try {
            basemap.mapPrefix("lx".intern(), LxReader.LXNAMESPACE); // default
            setProperty(LxReader.LX_GENERATED_PREFIX, "lxns");
        } catch (LxException e) {
            // Can't happen
            assert false;
        }
    }

    /** Returns true if the stack is empty */
    boolean empty() {
        return stacktop == null;
    }

    /** 
     * Push a new element on to the stack
     * @param elem the Element which is to be saved
     */
    void push(StackElement elem) {
        elem.setNext(stacktop);
        stacktop = elem;
    }
    
    /** Pop the top element off the stack */
    Promise pop() 
            throws LxException {
        if (stacktop == null)
            throw new LxException("pop: stack empty");
        Promise p = stacktop.getPromise();
        stacktop = stacktop.next();
        prefixnum--;
        return p;
    }
    
    /**
     * Retrieves the URI which a given prefix maps to.
     *
     * @param prefix the prefix to find the mapping of
     * @return the corresponding prefix-namespace pair, or null if there is
     * no such prefix mapping
     */
    PrefixNamespacePair getPrefixMapping(String prefix) {
        Object o = lookupKey(prefix);
        assert o == null || o instanceof PrefixNamespacePair;
        return (PrefixNamespacePair)o;
    }

    /**
     * Retrieves the value of a given feature.
     *
     * <p>If there is no mapping for the feature, we return false.
     * This should generally not happen, but it is the caller's
     * responsibility to avoid asking for such features, not ours.
     * Also, there are some cases, such as during initialisation,
     * when we ask for a feature we haven't initialised yet.
     *
     * @param feature the feature which we want the value of, which
     * must not be null
     * @return the value of the feature, or false if the feature is
     * unrecognised 
     * @throws IllegalArgumentException if the feature is null
     * @see #setProperty
     */
    boolean getFeature(String feature) 
            throws IllegalArgumentException {
        if (feature == null)
            throw new IllegalArgumentException("getFeature: null feature");
        Object o = lookupKey(feature);
        if (o == null)
            return false;
        assert o instanceof Boolean;
        return ((Boolean)o).booleanValue();
    }

    /**
     * Retrieves the value of a given property.
     *
     * <p>If there is no mapping for the property, we return false.
     * This should generally not happen, but it is the caller's
     * responsibility to avoid asking for such properties, not ours.
     * Also, there are some cases, such as during initialisation,
     * when we ask for a property we haven't initialised yet.
     *
     * @param property the property which we want the value of, which
     * must not be null
     * @return the value of the property, or null if the property is
     * unrecognised 
     * @throws IllegalArgumentException if the property is null
     * @see #setProperty
     */
    Object getProperty(String property) 
            throws IllegalArgumentException {
        if (property == null)
            throw new IllegalArgumentException("getProperty: null property");
        return lookupKey(property);
    }

    /**
     * Retrieves the attribute mapping of a given element
     * @return the list this element maps to, or null if there is no
     * such mapping
     */
    Stack.AttributeInfo getElementDecl(QName gi) {
        Object o = lookupKey(gi);
        assert o == null || o instanceof Stack.AttributeInfo;
        return (Stack.AttributeInfo)o;
    }

    /**
     * Sets a given prefix-namespace mapping in the stack.  If the
     * stack has any mappings in it, this is set in the topmost one;
     * if it has none, it is set in the base map.
     *
     * @param prefix the key, which may be null
     * @param uri the namespace it maps to, which may not be null
     * @throws LxException if the prefix starts with the prefix which is
     * currently mapped to the Lx namespace,
     * @throws IllegalArgumentException if the value is null
     */
    public void setPrefixMapping(String prefix, String uri)
            throws LxException {
        // There is an apparent namespace problem here, since this
        // map is used to store both prefix mappings and features.
        // However features are syntactically URIs (so including
        // colons), and prefixes are not (specifically not including
        // colons), so there's no danger of clashes in fact.
        if (uri == null)
            throw new IllegalArgumentException("setPrefixMapping: null uri");
        Mappings m = findTopmostMappings();
        if (m == null) {
            assert basemap != null;
            basemap.mapPrefix(prefix, uri);
        } else {
            m.mapPrefix(prefix, uri);
        }
    }

    /**
     * Does the prefix mapping, in the given map.  The given prefix
     * is mapped to a {@link #PrefixNamespacePair} which represents
     * the mapping.
     *
     * <p>The prefix may be null or the empty string, and in either
     * case a new prefix is generated, and it is the pair of the new
     * prefix and the namespace which the given prefix is mapped to.
     * The namespace may also be null, but if it is not null, it
     * should not be the empty string (this means the same as a null
     * namespace, but it is the responsibility of code elsewhere to do
     * the translation).  However it is an error for
     * <em>both</em> the prefix and the namespace to be null
     *
     * @param prefix the prefix which is to be mapped
     * @param ns the namespace it is to be mapped to
     * @param map the Map in which the mapping is to be made
     * @throws LxException if the above restrictions on the prefix
     * and namespace are not honoured
     */
    private void doPrefixMap(String prefix, String ns, Map map)
            throws LxException {

        String p;
        if (prefix == null && ns == null)
            throw new LxException("Can't map null prefix to null namespace");
        if (ns != null && ns.length() == 0)
            throw new LxException
                    ("doPrefixMap given empty namespace: should be null");
        if (map == null)
            throw new LxException("doPrefixMap called with null map"); 
        
        String ip = (prefix == null ? null : prefix.intern());
        String ins = (ns == null ? null : ns.intern());
        if (ins == LxReader.LXNAMESPACE) {
            if (ip == null || ip.length() == 0)
                // Don't allow this -- it would make the prefix
                // generation above impossible
                throw new LxException("Can't map null or empty prefix to "
                                      + LxReader.LXNAMESPACE);
            // This is a special case -- put in a reverse map
            map.put(LxReader.LXNAMESPACE, ip);
        }

        Object lxo = lookupKey(LxReader.LXNAMESPACE);
        assert lxo != null && lxo instanceof String;
        String lxPrefix = (String)lxo;
        if (ip == null || ip.length() == 0) {
            // generate a prefix
            if (prefixbase == null) {
                prefixbase = (String)getProperty(LxReader.LX_GENERATED_PREFIX);
                assert prefixbase != null;
            }
            if (prefixnum < 0) {
                prefixnum = 0;
                p = prefixbase;
            } else {
                prefixnum++;
                p = prefixbase + Integer.toString(prefixnum);
            }
        } else {
            if (prefix.startsWith(lxPrefix)
                && ins != LxReader.LXNAMESPACE)
                throw new LxException("Can't map prefix " + prefix
                                      + ": starts with reserved Lx prefix "
                                      + lxPrefix);
            p = ip;
        }

        assert p != null;
        assert ins == null || ins.length() > 0;
        map.put(ip, new PrefixNamespacePair(p, ins));
    }

    /**
     * Sets the value of a given feature in the stack.  If the
     * stack has any mappings within it, this is set in the topmost
     * mapping; if it has none, it is set in the base map.
     *
     * @param uri the feature to set, which may not be null
     * @param value the value to set the feature to
     * @throws IllegalArgumentException if the uri is null
     */
    public void setFeature(String uri, boolean value) {
        assert basemap != null;

        if (uri == null)
            throw new IllegalArgumentException
                    ("setFeature: null URI");
        uri = uri.intern();

        Mappings m = findTopmostMappings();
        (m == null ? basemap : m).mapFeature(uri, value);
    }

    /**
     * Sets the value of a given property in the stack.  If the
     * stack has any mappings within it, this is set in the topmost
     * mapping; if it has none, it is set in the base map.
     *
     * @param uri the property to set, which may not be null
     * @param value the value to set the property to
     * @throws IllegalArgumentException if the uri is null
     */
    public void setProperty(String uri, Object value) {
        assert basemap != null;

        if (uri == null)
            throw new IllegalArgumentException
                    ("setProperty: null URI");
        uri = uri.intern();

        Mappings m = findTopmostMappings();
        (m == null ? basemap : m).mapProperty(uri, value);
    }

    /**
     * Looks up the given key in the stack and in the base map
     * @param key the key to be searched for
     * @return the key value, or null if none found
     */
    private Object lookupKey(Object key) {
        Object v = stacktop == null ? null : stacktop.lookup(key);
        if (v == null)
            v = basemap.lookup(key);
        return v;
    }

    /**
     * Finds the topmost stack frame which is a Mappings element.
     * @return a Mappings element, or null if there are none in the stack
     */
    private Mappings findTopmostMappings() {
        for (StackElement se = stacktop;
             se != null;
             se = se.next())
            if (se instanceof Mappings)
                return (Mappings)se;
        return null;
    }

    /**
     * A fragment of code which will be stored up and run later.
     */
    public static abstract class Promise {
        private Promise next;
        /**
         * The work to be done.  Classes which extend this class
         * should implement this method.
         */
        protected abstract void doit();
        /**
         * Force the promise.  When it comes time to call in the
         * promise, this method is called on the Promise object.
         * This Promise's <code>doit</code> method, and the
         * <code>doit</code> methods of any promises appended to this
         * one, will be invoked.
         * @see #appendPromise
         */
        final public void force() {
            doit();
            if (next != null)
                next.force();
        }
        /**
         * Append another Promise to be run after this one.
         * @param promise another promise to be run after this one
         */
        final public void appendPromise(Promise promise) {
            if (next == null)
                next = promise;
            else
                next.appendPromise(promise);
        }
    }
    

    /**
     * An element of the stack.  One of the properties which a stack
     * can have is a `Promise', which is what is returned when the
     * stack is popped.
     */
    private static class StackElement {
        private Promise promise;
        private StackElement next;
        
        StackElement() {
            this(null);
        }
        StackElement(Promise promise) {
            this.promise = promise;
            this.next = null;
        }
        void setNext(StackElement se) {
            assert next == null;
            next = se;
        }
        Promise getPromise() {
            return promise;
        }
        Object addPromise(Promise p) {
            Promise oldp = promise;
            if (promise == null)
                promise = p;
            else
                promise.appendPromise(p);
            return oldp;
        }
        /**
         * Lookup a key in the stack.  This queries the current stack
         * element (if there are any mappings in this element) and all
         * those below it in the stack.
         *
         * @param key the value to be queried, which may be null
         * @return the value corresponding to the key, or null if no
         * mapping was found
         */
        Object lookup(Object key) {
            return next() == null ? null : next().lookup(key);
        }
        StackElement next() {
            return next;
        }
        public String toString() {
            return "null-StackElement!";
        }
    }

    static class Element
            extends StackElement {
        String elementName;
        Element(String el, Promise promise) {
            super(promise);
            elementName = el.intern();
        }
        String getElementName() {
            return elementName;
        }
        public String toString() {
            return "Element(" + elementName + ")";
        }
    }

    Mappings newMapping() {
        return new Mappings();
    }

    /**
     * A collections of one or mare mappings.  This can map both
     * prefixes and features.
     */
    class Mappings 
            extends StackElement {
        private Map map;

        private Mappings() {
            super(null);
        }

        /**
         * Map a prefix to a URI.
         *
         * <p>The prefix may be null or the empty string, and in either
         * case a new prefix is generated, and it is the pair of the new
         * prefix and the namespace which the given prefix is mapped to.
         * The namespace may also be null, but if it is not null, it
         * should not be the empty string (this means the same as a null
         * namespace).  However it is an error for
         * <em>both</em> the prefix and the namespace to be null
         *
         * @param prefix the prefix which is to be mapped
         * @param ns the namespace it is to be mapped to
         * @throws LxException if the above restrictions on the prefix
         * and namespace are not honoured
         */
        void mapPrefix(String prefix, String ns) 
                throws LxException {
            if (map == null)
                map = makeMap();

            doPrefixMap(prefix, ns, map);
        }

        void mapFeature(String uri, boolean value) {
            if (uri == null)
                throw new IllegalArgumentException
                        ("Stack.mapFeature received null URI");
            if (map == null)
                map = makeMap();
            map.put(uri.intern(), Boolean.valueOf(value));
        }

        void mapProperty(String uri, Object value) {
            if (uri == null)
                throw new IllegalArgumentException
                        ("Stack.mapProperty received null URI");
            if (map == null)
                map = makeMap();
            map.put(uri.intern(), value);
        }

        void mapElementDecl(QName gi, LxList attlist) 
                throws LxException {
            AttributeInfo ai;

            if (attlist.length() == 1 && !attlist.car().isTrue()) {

                // The attlist is just (#f) -- the gi has unrestricted content
                ai = new AttributeInfo(false);
                
            } else {

                ai = new AttributeInfo();
                for (; !attlist.isNull(); attlist=attlist.cdr()) {
                    LxObject a = attlist.car();
                    
                    if (a.isStringLike()) {

                        ai.addAttribute(new QName(a.toString(),
                                                  Stack.this,
                                                  gi),
                                        false);

                    } else if (a.isList()) {

                        LxList pair = (LxList)a;
                        if (pair.length() < 1 || pair.length() > 2)
                            throw new LxException
                                    ("Dtd pair " + pair + " wrong length");
                        ai.addAttribute
                                (new QName(pair.car().toString(),
                                           Stack.this,
                                           gi),
                                 pair.length() == 2
                                 && pair.cdr().car().isTrue());

                    } else {
                        throw new LxException("Dtd spec " + a
                                              + " has wrong type");
                    }
                }
            }

            if (map == null)
                map = makeMap();
            map.put(gi, ai);
        }

        private Map makeMap() {
            // Return a new LinkedHashMap: order is predictable,
            // but it accepts nulls as keys, unlike TreeMap.  This is
            // not vital to the functionality of the stack, but (a)
            // it helps testing, and (b) it makes things easier
            // for the eventual clients.
            return new java.util.LinkedHashMap(4);
        }

        Object lookup(Object key) {
            if (map != null && map.containsKey(key))
                return map.get(key);
            else
                // Can add caching here
                return super.lookup(key);
        }
        
        public Iterator getKeyIterator() {
            if (map == null)
                return new Iterator() {
                        public boolean hasNext() {
                            return false;
                        }
                        public Object next() 
                                throws java.util.NoSuchElementException {
                            throw new java.util.NoSuchElementException
                                    ("No such element");
                        }
                        public void remove()
                                throws UnsupportedOperationException {
                            throw new UnsupportedOperationException 
                                    ("No operation remove()");
                        }
                    };
            else
                return map.keySet().iterator();
        }
        
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append("Mapping(");
            if (map != null) {
                for (Iterator i = map.entrySet().iterator();
                     i.hasNext();
                     ) {
                    Map.Entry me = (Map.Entry)i.next();
                    sb.append(me.getKey().toString())
                            .append("=")
                            .append(me.getValue().toString())
                            .append(',');
                }
            }
            sb.append(")");
            return sb.toString();
        }
    }

    class AttributeInfo {
        private java.util.List attlist;
        private boolean restricted;
        private class OneAttribute {
            final QName q;
            final boolean v;
            OneAttribute(QName qname, boolean value) {
                q = qname;
                v = value;
            }
        }

        private AttributeInfo() {
            this(true);
        }

        private AttributeInfo(boolean restricted) {
            attlist = new java.util.ArrayList();
            this.restricted = restricted;
        }
            
        private void addAttribute(QName att, boolean required) {
            // we mustn't call this on an object which has been
            // created unrestricted
            assert restricted;
            attlist.add(new OneAttribute(att, required));
        }
        private OneAttribute getatt(int i) {
            return (OneAttribute)attlist.get(i);
        }
        public boolean allowedAttribute(QName qn) {
            if (restricted) {
                for (int i=0; i<attlist.size(); i++)
                    if (getatt(i).q.equals(qn))
                        return true;
                return false;
            } else {
                return true;
            }
        }
        public Iterator allRequired() {
            int i;
            for (i=0; i<attlist.size(); i++)
                if (getatt(i).v)
                    break;
            final int first = i;
            return new Iterator() {
                    private int nexti = first;
                    public boolean hasNext() {
                        return nexti < attlist.size();
                    }
                    public Object next() 
                            throws java.util.NoSuchElementException {
                        if (nexti >= attlist.size())
                            throw new java.util.NoSuchElementException();
                        Object o = getatt(nexti).q;
                        for (nexti++; nexti<attlist.size(); nexti++)
                            if (getatt(nexti).v)
                                break;
                        return o;
                    }
                    public void remove()
                            throws UnsupportedOperationException {
                        throw new UnsupportedOperationException 
                                ("No operation remove()");
                    }
                };
        }
    }

    /**
     * A prefix-Namespace pair.  The prefix, but not the namespace
     * URI, is guaranteed to be non-null.  The strings returned by the
     * methods are interned using {@link java.lang.String#intern}.
     */
    public static class PrefixNamespacePair {
        private String prefix;
        private String ns;
        PrefixNamespacePair(String prefix, String ns) {
            this.prefix = prefix.intern();
            this.ns = (ns == null ? null : ns.intern());
            assert this.prefix != null;
        }
        /** 
         * Returns the prefix.
         * @return the prefix of this pair, non-null
         */
        public String getPrefix() {
            assert prefix != null;
            return prefix;
        }
        /**
         * Returns the namespace this prefix is mapped to
         * @return the namespace as a URI, or null if this prefix is mapped to
         * a null namespace
         */
        public String getNamespace() {
            return ns;
        }
        public String toString() {
            return prefix + ">" + ns;
        }
    }
}
