// This file is part of Lx.
//
// Copyright 2003, 2004, 2008, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.tokenizer;

import uk.me.nxg.lx.LxException;

import java.io.*;

/**
 * A class which tokenises an input file.  Tokens are obtained by
 * calling the getToken() method.
 *
 * <p>The Tokenizer returns tokens in the following classes
 * <dl>
 * <dt>Token.OPENPAREN and Token.CLOSEPAREN
 * <dd>The tokens which open and close lists
 *
 * <dt>SYMBOL
 * <dd>A bare string, such as <code>word</code>, which does not fall
 * into one of the other categories.  It may contain any characters
 * other than whitespace and '(' and ')'.
 *
 * <dt>STRING
 * <dd>A sequence of characters, delimited by either
 * <code>"..."</code> or <code>[...]</code>.  In both cases, a
 * backslash may escape any character.  The <code>"..."</code> form
 * may <em>not</em> include newline or carriage-return characters.
 * The <code>[...]</code> form may include newlines, and may include
 * both quotes, and other '[' or ']' characters as long as they are
 * properly nested.  Thus all of <code>"hell\"o"</code>,
 * <code>[h[ell]o]</code> and <code>[hel\]lo]</code> are legal strings.
 *
 * <dt>TOKEN
 * <dd>A string starting with <code>#</code>, such as <code>#f</code>
 *
 * <dt>KEYWORD
 * <dd>A string ending with a colon, such as
 * <code>attribute:</code>.  Colons elsewhere in the string are not
 * recognised by this lexer, though they are significant elsewhere.
 *
 * <dt>EOF
 * <dd>The end-of-file token, returned when the Tokenizer has read
 * beyond the last character of the input
 *
 * </dl>
 *
 * <p>Each of these tokens has a symbolic constant of the same name.
 * There is also a <code>BAD_TOKEN</code> token, which is never returned,
 * but used for assertion checking internally.
 *
 * <p>After the Tokenizer has returned one of the token types
 * <code>SYMBOL</code>, <code>STRING</code>, <code>TOKEN</code> or
 * <code>KEYWORD</code>, the value of the token (minus the leading
 * <code>#</code> or trailing colon in the case of the latter two
 * types) is available by calling the method {@link #value}.
 *
 * <p>The Tokenizer has two tokens-worth of pushback -- see {@link #ungetToken}.
 *
 * <p>The class also implements the SAX <code>Locator</code>
 * interface, so (as well as being useful itself) it may be used as a
 * SAX locator in the constructor of a SAX exception.
 *
 * @see #getToken
 *
 * @author Norman Gray
 * @version 0.5.3, 2012 August 5
 */
public class Tokenizer
	implements org.xml.sax.Locator {
    public enum Token {
        BAD_TOKEN, EOF, SYMBOL, STRING, TOKEN, KEYWORD, OPENPAREN, CLOSEPAREN }
    
    private boolean traceTokens = false;

    // Newline characters
    static final private char CHAR_LF = '\n';
    static final private char CHAR_CR = '\r';
    static final private char CHAR_NEL = '\u0085';

    /** The value of the string which was last read from the input. */
    private String currentString;

    private PushbackReader i_;
    private LineNumberReader r_;
    private String sysid;

    /** The name of the file we're reading from.  Null if none known. */
    private String filename;

    /** 
     * The last character read from the input.  This is almost private
     * to {@link #getToken}.
     */
    private int ch;

    /** The (short) arrays of pushed-back tokens and strings.
     * We need two tokens of pushback. */
    private Token pushbackTokens[] = { Token.BAD_TOKEN, Token.BAD_TOKEN };
    private String pushbackStrings[] = { null, null };
    /** The number of tokens in the pushback buffer.  Note this is one
     * greater than the index of the next token to be retrieved.  */
    private int nPushbackTokens = 0;


    /** Constructs a tokenizer from a File object.
     *
     * @param f The file to be opened
     * @exception java.io.FileNotFoundException when there is no input file
     * @exception java.io.IOException on other IO errors
     */
    public Tokenizer (File f)
	throws FileNotFoundException, IOException
    {
	r_ = new LineNumberReader (new FileReader(f));
	i_ = new PushbackReader (r_);
        filename = f.getName();
	try {
	    sysid = f.toURL().toExternalForm();
	} catch (java.net.MalformedURLException e) {
	    throw new FileNotFoundException
		    ("Can't get SystemID for file: " + e);
	}

        assert i_ != null;
        assert sysid != null;
    }

    /**
     * Constructs a tokenizer from an InputStream
     *
     * @param reader the source to read from
     * @throws IOException if the reader is null
     */
    public Tokenizer (Reader reader) 
	    throws IOException {
	if (reader == null)
	    throw new IOException("Tokenizer initialised with null reader");

	r_ = new LineNumberReader(reader);
	i_ = new PushbackReader(r_);
	sysid = "reader:" + reader.toString(); // not very good!

        assert i_ != null;
        assert sysid != null;
    }

    /** Constructs a tokenizer from a String
     *
     * @param s The string to be parsed
     */
    public Tokenizer (String s)
    {
	r_ = null;
	i_ = new PushbackReader (new StringReader (s));
	sysid = "string:" + s;	// not ideal

        assert i_ != null;
        assert sysid != null;
    }

    /** Closes the input stream 
     * @exception IOException on IO errors
     */
     public void close ()
	throws IOException
    {
        if (r_ != null) {
            r_.close();
            r_ = null;
        }
	if (i_ != null) {
	    i_.close();
	    i_ = null;
	}
    }

    private static boolean isEOL (int c)
    {
	// I'd have thought there was a more unicode-ish way of doing this,
	// but Character.getType() == Character.LINE_SEPARATOR apparently
	// isn't it
	return c == '\n' || c == '\r';
    }

    /**
     * Returns a token from the input stream.  Returned tokens are:
     * <dl>
     * <dt>Token.SYMBOL<dd>A bare string.  This is indicated by
     * either a bare word or a word preceded by a single quote (any
     * trailing single quote is taken to be part of the symbol).
     * <dt>Token.TOKEN<dd>A symbol starting with `#', minus the
     * `#' character. 
     * <dt>Token.KEYWORD<dd>A symbol ending with `:', minus the
     * `:' character. 
     * <dt>Token.STRING<dd>String enclosed in double quotes.
     * Double quotes can be included in the string by escaping them as
     * <code>\"</code>.
     * <dt>Token.OPENPAREN or Token.CLOSEPAREN<dd>Start and end of lists.
     * <dt>Token.EOF<dd>End of file.
     * </dl>
     * The first four return types make the read string available via
     * the method {@link #value}.
     *
     * @return the next token
     * @throws IOException on IO errors
     * @throws LxException on tokenization errors, such as unexpected EOF
     */
    public Token getToken ()
	    throws IOException, LxException
    {
	Token result;
	StringBuffer buf;

        if (nPushbackTokens > 0) {
            assert nPushbackTokens <= pushbackTokens.length;
            nPushbackTokens--;
            currentString = pushbackStrings[nPushbackTokens];
            if (traceTokens)
                System.err.println
                        ("getToken-->"
                         + tokenToString(pushbackTokens[nPushbackTokens]));
            result = pushbackTokens[nPushbackTokens];
            assert result != Token.BAD_TOKEN;
            return result;
        }

        if (i_ == null)         // close() has been called
            return Token.EOF;

	// skip whitespace and comments
	do 
	{
	    ch = i_.read();
	    if (ch == ';')
		do
		    ch = i_.read();
		while (!isEOL(ch));
	}
        while (ch != -1 && Character.isWhitespace((char)ch));

	switch (ch)
	{
	  case -1:
            close();            // sets i_ = null
	    result = Token.EOF;
            currentString = null;
	    break;

	  case '(':
            result = Token.OPENPAREN;
            currentString = null;
            break;
          case ')':
            result = Token.CLOSEPAREN;
            currentString = null;
	    break;

	  case '\"':
	    // Read in a string delimited by "...",
	    // allowing some escapes (\", \\, \n, \t).  Newlines are OK.
	    result = Token.STRING;
	    buf = new StringBuffer ();
	    while ((ch = i_.read()) != '\"') {
		if (ch == '\\') {
		    ch = i_.read();
                    switch (ch) {
                      case '\\':
                      case '\"':
                      case -1:
                        // do nothing
                        break;
                      case 'n':
                        ch = '\n';
                        break;
                      case 'r':
                        ch = '\r';
                        break;
                      case 't':
                        ch = '\t';
                        break;
                      default:
                        System.err.println("Unexpected escape \\" + (char)ch
                                           + " reading string");
                        break;
                    }
                }
		if (ch == -1) {	// EOF
                    close();
		    throw new LxException("Unexpected EOF reading string");
                }
		buf.append((char)ch);
	    }
	    currentString = buf.toString();
	    break;

          case '[':
            // Read in a string delimited by [...], allowing
            // backslashes to quote, and allowing nested square
            // brackets.  Also allow newlines.
            {
                result = Token.STRING;
                buf = new StringBuffer ();
                int nesting = 1;
                boolean escaped;
                while (nesting > 0) {
                    ch = i_.read();
                    escaped = (ch == '\\');
                    if (escaped)
                        ch = i_.read();
                    if (ch == -1) { // EOF
                        close();
                        throw new LxException("unexpected EOF reading string");
                    }
                    if (!escaped) {
                        if (ch == '[')
                            nesting++;
                        else if (ch == ']')
                            nesting--;
                    }
                    if (nesting > 0)
                        buf.append((char)ch);
                }
                currentString = buf.toString();
            }
            break;

          default:
	    if (ch == '#') {
		result = Token.TOKEN;
		// Discard the # character
		ch = i_.read();
	    } else if (ch == '\'') {
		// Categorise this as a symbol, and discard the quote
		// (ie, we don't have any lisp-like quoting, as such).
		result = Token.SYMBOL;
		ch = i_.read();
	    } else {
		// a bare string is a symbol, just like the quoted string above
		result = Token.SYMBOL;
	    }
            if (ch == -1)
                throw new LxException("Unexpected EOF reading characters");

	    buf = new StringBuffer();
	    buf.append ((char)ch);
	    while (! ((ch = i_.read()) == -1
                      || Character.isWhitespace((char)ch)
                      || ch == '('
                      || ch == ')'
                      || ch == '\"'
                      || ch == '['
                      || ch == ']'
                      || ch == ';'))
		buf.append ((char)ch);
            if (ch == -1) {
                /* Avoid doing unread(-1).  In JDK 1.4.1 at least,
                 * after unread(-1), read() returns 65535. */
                close();        // sets i_ = null
                assert i_ == null;
            } else {
                i_.unread (ch);
            }
	    if (result == Token.SYMBOL && buf.charAt(buf.length()-1) == ':')
	    {
		// Ah, it turns out that this SYMBOL was really a KEYWORD
		result = Token.KEYWORD;
		buf.setLength(buf.length()-1); // lose the colon
	    }
	    currentString = buf.toString();
	    break;
	}

	assert result != Token.BAD_TOKEN;

        if (traceTokens)
            System.err.println
                    ("getToken->"
                     + tokenToString(pushbackTokens[nPushbackTokens]));

        return result;
    }

    /**
     * Push the given token back into the tokenizer.  This token will
     * be the next one to be returned by {@link #getToken}.  You may
     * push back a different token type from the one last obtained
     * from <code>getToken</code>, but you cannot change the string
     * value of the token as you do so.
     *
     * <p>You can push back a maximum of two tokens.
     *
     * @param token one of the token types
     * @throws LxException if there are already two tokens pushed
     * back, or if the token is not a valid one
     */
    public void ungetToken(Token token) 
	    throws LxException {
        ungetToken(token, null);    
    }

    /**
     * Push the given token and string back into the Tokenizer.  This
     * is as with {@link #ungetToken(Token)}, but allows you to change
     * the string which will be returned.
     *
     * <p>You can push back a maximum of two tokens.
     *
     * @param token one of the token types
     * @param s the string associated with the token
     * @throws LxException if there are already two tokens pushed
     * back, or if the token is not a valid one
     */
    public void ungetToken(Token token, String s)
            throws LxException {
        if (nPushbackTokens >= pushbackTokens.length) {
	    throw new LxException
		    ("Too many tokens pushed back in Tokenizer");
	}
//         // check this is a valid token
//         if (! okToken(token)) {
//             throw new LxException("Attempt to push back invalid token "
//                                   + token, this);
//         }

        pushbackTokens[nPushbackTokens] = token;
        pushbackStrings[nPushbackTokens] = (s == null ? currentString : s);
        nPushbackTokens++;
        if (traceTokens)
            System.err.println("ungetToken<-" + tokenToString(token));
    }

    /**
     * Returns the value of the token as a string.  If the last token
     * was one of the token types <code>SYMBOL</code>,
     * <code>STRING</code>, <code>TOKEN</code> or
     * <code>KEYWORD</code>, then this is the string content of the
     * token.  If the last token was a TOKEN or KEYWORD, then the hash
     * or colon character (respectively) is removed before return.  If
     * the last token was one of the other types, then the return
     * value of this method is guaranteed to be non-null, but the
     * actual value is undefined.
     *
     * @return the string value of the last token
     */
    public String value()
    {
	String ret = (currentString == null ? "<none>" : currentString);
        assert ret != null;
        return ret;
    }

    /**
     * Returns the current line number in the input file.
     * The first line in the file is line number 1 (rather than starting
     * at zero).  The number is always 1 if we are Tokenizing a string),
     * and always positive
     * @return the current line number
     */
    public int getLineNumber()
    {
        int currLineNo;
	if (r_ == null)
	    currLineNo = 1;
	else {
            currLineNo = r_.getLineNumber();
            // r_.getLineNumber is zero offset, so increment it unless we're
            // currently at the end of a line.  In this case r_.getLineNumber()
            // refers to the _next_ line, so by not incrementing it,
            // it holds what we regard as the correct line number.
            if (!isEOL(ch))
                currLineNo++;
        }
        assert currLineNo > 0;
	return currLineNo;
    }

    /**
     * Returns the current column number.  Always returns -1 in fact.
     * @return -1
     */
    public int getColumnNumber() {
	return -1;
    }

    /**
     * Obtains a system ID for the object being tokenized.
     * @return a system ID
     */
    public String getSystemId() {
        assert sysid != null;
        return sysid;
    }

    /**
     * Sets the system ID for the object being tokenized.  This is
     * useful when the system ID cannot realistically be deduced
     * within the constructor (for example, when the constructor was
     * called with a <code>reader</code>).
     *
     * @param systemId a system ID for this object;
     * if null, the previous (possibly default) system ID is unchanged
     */
    public void setSystemId(String systemId) {
        if (systemId != null)
            sysid = systemId;
        assert sysid != null : "Null sysid after setting";
    }

    /**
     * Obtains a public ID for the object being tokenized.  Always
     * returns null in fact. 
     * @return null
     */
    public String getPublicId() {
        return null;
    }

    /** Returns the name of the current input file.
     * @return filename of the current input file, or "-" if the class
     * is tokenising a String.
     */
    public String getFilename ()
    {
        return (filename == null ? "-" : filename);
    }

    /**
     * Translate a token to a string, for debugging purposes.  The
     * string returned is accurate only if the token queried is the
     * last token returned.
     * @param token the token to be converted, which should be the
     * last token returned from the Tokenizer
     * @return a String representation of the token
     */
    public String tokenToString(Token token)
            throws LxException {
        return token.toString();
    }

    public void traceTokenization(boolean flag) {
        traceTokens = flag;
    }

    /** Finalises the object
     * @exception java.lang.Throwable from superclass?
     */
    protected void finalize ()
	throws Throwable
    {
	super.finalize();
	close();
    }	
}
