// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;
import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

import junit.framework.Assert;
import junit.framework.AssertionFailedError;
import junit.framework.TestCase;

public class LxObjectTest
        extends TestCase {

    static public final int LIST = 1;
    static public final int STRING = 2;
    static public final int SYMBOL = 3;
    static public final int TOKEN = 4;
    static public final int KEYWORD = 5;

    public LxObjectTest(String name) {
        super(name);
    }

    public void testRead()
            throws Exception {

        LxObject o, o2;
        LxList l, l2;

        o = getObject("");
        assertNull(o);

        o = getObject("hello");
        assertNotNull(o);
        assertDisjointTypes(o, SYMBOL);
        assertEquals("hello", o.toString());

        o = getObject("\"hello\"");
        assertNotNull(o);
        assertDisjointTypes(o, STRING);
        assertEquals("hello", o.toString());

        o = getObject("hello:");
        assertNotNull(o);
        assertDisjointTypes(o, KEYWORD);
        assertEquals("hello", o.toString());

        o = getObject("#hello");
        assertNotNull(o);
        assertDisjointTypes(o, TOKEN);
        assertEquals("hello", o.toString());

        o = getObject("()");
        assertNotNull(o);
        assertDisjointTypes(o, LIST);
        assertTrue(o instanceof LxList);
        l = (LxList)o;
        assertTrue(l.isList());
        assertTrue(l.isNull());
        assertNull(l.car());


        o = getObject("(hello)");
        assertNotNull(o);
        assertDisjointTypes(o, LIST);
        assertTrue(o instanceof LxList);
        l = (LxList)o;
        assertFalse(l.isNull());
        assertEquals(1, l.length());
        o2 = l.car();
        assertNotNull(o2);
        assertDisjointTypes(o2, SYMBOL);
        assertEquals("hello", o2.toString());
        assertNotNull(l.cdr());
        l = l.cdr();            // empty list
        assertDisjointTypes(l, LIST);
        assertTrue(o instanceof LxList);
        //l = (LxList)o;
        assertTrue(l.isNull());


        // Make sure ( does terminate symbols
        Tokenizer t = new Tokenizer("hello()");
        o = getObject(t);
        assertNotNull(o);
        assertDisjointTypes(o, SYMBOL);
        assertEquals("hello", o.toString());
        o = getObject(t);
        assertNotNull(o);
        assertDisjointTypes(o, LIST);
        assertTrue(((LxList)o).isNull());
        assertEquals(Token.EOF, t.getToken());
        

        // More complicated list
        o = getObject("(\"hello\" (there now:))");
        assertNotNull(o);
        assertDisjointTypes(o, LIST);
        assertTrue(o instanceof LxList);
        l = (LxList)o;
        assertFalse(l.isNull());
        assertEquals(2, l.length());
        o2 = l.car();  // "hello"
        assertNotNull(o2);
        assertDisjointTypes(o2, STRING);
        assertEquals("hello", o2.toString());
        // examine rest of list
        assertNotNull(l.cdr());
        o = l.cdr();            // ((there now:))
        assertDisjointTypes(o, LIST);
        l = (LxList)o;
        o2 = l.car();           // (there now:)
        assertDisjointTypes(o2, LIST);
        l2 = (LxList)o2;
        assertEquals(2, l2.length());
        assertDisjointTypes(l2.car(), SYMBOL);
        assertEquals("there", l2.car().toString());
        assertDisjointTypes(l2.cdr(), LIST);
        l2 = (LxList)l2.cdr();          // (now:)
        assertDisjointTypes(l2.car(), KEYWORD);
        assertEquals("now", l2.car().toString());
        assertDisjointTypes(l2.cdr(), LIST);
        l2 = (LxList)l2.cdr();  // empty list
        assertDisjointTypes(l2, LIST);
        assertTrue(l2.isNull());
        // examine rest of list
        l = (LxList)l.cdr();
        assertDisjointTypes(l, LIST);
        assertEquals(0, l.length());
        assertTrue(l.isNull());
        

        // The following is not an error -- we simply read nothing
        // from the token stream
        t = new Tokenizer(")");
        o = getObject(t);
        assertNull(o);
        assertEquals(Token.CLOSEPAREN, t.getToken());
        assertEquals(Token.EOF, t.getToken());

        // Likewise
        t = new Tokenizer("hello)");
        o = getObject(t);
        assertNotNull(o);
        assertDisjointTypes(o, SYMBOL);
        assertEquals(Token.CLOSEPAREN, t.getToken());
        assertEquals(Token.EOF, t.getToken());
    }

//     public void testListOperations()
//             throws Exception {
//         LxObject s = getObject("s");
//         LxObject emptyl = getObject("()");
//         LxObject l = getObject("(e1)");

//         assertDisjointTypes(s, STRING);
//         assertDisjointTypes(emptyl, LIST);
//         assertDisjointTypes(l, LIST);

//         assertEquals(0, emptyl.length());
//         assertEquals(1, l.length());

//         l = l.append(emptyl);
//         assertDisjointTypes(l, LIST);
//         assertEquals(1, l.length());

//         l = l.append(l);
//         assertDisjointTypes(l, LIST);
//         assertEquals(2, l.length());
//         assertTrue(l.cdr().car().isString());
//         assertEquals("e1", l.cdr().car().toString());

//         l = l.append(s);
//         assertDisjointTypes(l, LIST);
//         assertEquals(3, l.length());
//         assertTrue(l.cdr().car().isString());
//         assertEquals("e1", l.cdr().car().toString());
//         assertTrue(l.cdr().cdr().car().isString());
//         assertEquals("s", l.cdr().cdr().car().toString());

//         l = LxList.EMPTYLIST;
//         assertDisjointTypes(l, LIST);
//         assertEquals(0, l.length());
//         l = l.append(s);
//         assertDisjointTypes(l, LIST);
//         assertEquals(1, l.length());
//         assertEquals("s", l.car().toString());
//     }
    
    public void testFailures()
            throws Exception {
        assertFailure("(");
        assertFailure("(hello there");
        assertFailure("(hello()");
    }

    static private LxObject getObject(Tokenizer t) 
            throws LxException {
        return LxObject.readLxObject(t);
    }
    
    static private LxObject getObject(String s) 
            throws LxException {
        return getObject(new Tokenizer(s));
    }

    /** Assert that a condition is false.  Older versions of JUnit don't have this. */
    static public void assertFalse(boolean test) {
        assertTrue(! test);
    }

    static private void assertDisjointTypes(LxObject o, int type)
            throws Exception {
        assertTrue((type==LIST)    == o.isList());
        assertTrue((type==STRING)  == o.isString());
        assertTrue((type==SYMBOL)  == o.isSymbol());
        assertTrue((type==TOKEN)   == o.isToken());
        assertTrue((type==KEYWORD) == o.isKeyword());
    }

    static private void assertFailure(String s) {
        try {
            LxObject o = getObject(s);
            fail("LxObject didn't fail with \"" + s + "\"");
        } catch (LxException e) {
            // OK
        } catch (Exception e) {
            fail("LxObject produced unexpected exception for input \""
                 + s + "\": " + e);
        }
    }
}
