// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;

import junit.framework.Assert;
import junit.framework.TestCase;


public class QNameTest 
	extends TestCase {

    public QNameTest(String name) {
	super(name);
    }

    public void testConstructor()
            throws Exception {

        Stack s = new Stack();

        Stack.Mappings pm = s.newMapping();
        try {
            // This is case 1, but real code should never call this,
            // but instead trap this case elsewhere
            pm.mapPrefix(null, null);
            fail("Mapping null prefix to null namespace did not fail");
        } catch (LxException e) {
            // OK
        }
        pm.mapPrefix("",   null);	// 3
        pm.mapPrefix("x1", null);	// 5
        s.push(pm);

        QName qn = new QName("gi", s);
        assertQName(qn, "gi", null, null, "gi");

        qn = new QName(":gi", s);
        assertQName(qn, "gi", null, null, "gi");

        qn = new QName("x1:gi", s);
        assertQName(qn, "gi", null, null, "gi");

        s.pop();

        
        pm = s.newMapping();
        pm.mapPrefix(null, "ns:1");	// case 2
        pm.mapPrefix("",   "ns:2");	// 4
        pm.mapPrefix("x2", "ns:3");	// 6
        s.push(pm);

        // The prefix names that we test here are not part of the
        // documented interface
        qn = new QName("gi", s);
        String testPrefix = qn.prefix();
        assertNotNull(testPrefix);
        String testQname = testPrefix + ":gi";
        //assertQName(qn, "lxns1:gi", "lxns1", "ns:1", "gi");
        assertQName(qn, testQname, testPrefix, "ns:1", "gi");

        qn = new QName(":gi", s);
        testPrefix = qn.prefix();
        assertNotNull(testPrefix);
        testQname = testPrefix + ":gi";
        //assertQName(qn, "lxns2:gi", "lxns2", "ns:2", "gi");
        assertQName(qn, testQname, testPrefix, "ns:2", "gi");

        qn = new QName("x2:gi", s);
        assertQName(qn, "x2:gi", "x2", "ns:3", "gi");

        // The following qnames should fail
        String[] badQNames = {
            "",                 // empty string
            " ",                // illegal character
            "#",                // ...
            ":",                // empty localname
            "x2:",              // ...
            "2:hello",          // prefix does not start with (Letter|'_')
            " :hello",          // ...
            "x2:4ello",         // localname does not start with (Letter|'_')
            "a:b:local",        // two colons
            "x2:a:local",       // ...ditto, but with an OK prefix
            "space char",       // has a space in it
            " something",       // ...
            "something ",       // ...
            "pref:name",        // unknown prefix
        };
        for (int i=0; i<badQNames.length; i++) {
            try {
                QName tqn = new QName(badQNames[i], s);
                fail("Creating QName with name " + badQNames[i]
                     + " did not fail");
            } catch (LxException e) {
                // OK
            }
        }

        s.pop();
    }

    public void testNSConstructor()
            throws Exception {
        QName qn = QName.newNSAttribute(null);
        assertQName(qn, "xmlns", null, null, "xmlns");
        qn = QName.newNSAttribute("hello");
        assertQName(qn, "xmlns:hello", null, null, "xmlns:hello");

        // Following should fail
        String[] bad = {
            "",                 // bad prefix
            " ",                // space
            "#",                // includes illegal character
            "hello#",           // ...
            "2x",               // prefix starts with illegal character
            " one",             // space
            "one ",             // ...
            "bad:prefix",       // includes colon
        };
        for (int i=0; i<bad.length; i++) {
            try {
                QName tqn = QName.newNSAttribute(bad[i]);
                fail("Creating NSAttribute with prefix " + bad[i]
                     + " did not fail");
            } catch (LxException e) {
                // OK
            }
        }
    }
    
    public void testEquals()
            throws Exception {

        Stack s = new Stack();

        Stack.Mappings pm = s.newMapping();
        pm.mapPrefix("a", "url:1");
        pm.mapPrefix("b", "url:1");
        s.push(pm);

        QName qn0 = new QName("gi", s);
        QName qn1 = new QName("a:gi", s);
        QName qn2 = new QName("b:gi", s);
        assertFalse(qn0.equals("gi"));
        assertTrue(qn0.hashCode() != "gi".hashCode());
        assertFalse(qn1 == qn2);
        assertTrue(qn1.equals(qn2));
        assertTrue(qn1.hashCode() == qn2.hashCode());
    }

    private void assertQName(QName q,
                             String qname,
                             String prefix,
                             String uri,
                             String localname) {
        assertNotNull(q.qName());
        assertEquals(qname, q.qName());
        if (prefix == null)
            assertNull(q.prefix());
        else
            assertEquals(prefix, q.prefix());
        if (uri == null)
            assertNull(q.namespace());
        else
            assertEquals(uri, q.namespace());
        assertNotNull(q.localName());
        assertEquals(localname, q.localName());
    }

    /** 
     * Assert that a condition is false.
     * Older versions of JUnit don't have this. 
     */
    static public void assertFalse(boolean test) {
        assertTrue(! test);
    }
}
