// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.reader;

import uk.me.nxg.lx.LxException;
import uk.me.nxg.lx.tokenizer.Tokenizer;

import junit.framework.Assert;
import junit.framework.TestCase;

import org.xml.sax.SAXNotRecognizedException;


public class StackTest
        extends TestCase {

    // for the makePromise helper method
    private String promiseKept = null;

    public StackTest(String name) {
        super(name);
    }

    public void testStacking()
            throws Exception {

        Stack s = new Stack();

        assertTrue(s.empty());
        s.push(new Stack.Element("one", makePromise("a")));
        assertFalse(s.empty());
        s.push(new Stack.Element("two", makePromise("b")));
        Stack.Promise p = s.pop();
        assertNotNull(p);
        p.force();
        assertEquals("b", promiseKept);
        s.push(new Stack.Element("three", makePromise("c")));
        p = s.pop();
        assertNotNull(p);
        p.force();
        assertEquals("c", promiseKept);
        p = s.pop();
        assertNotNull(p);
        p.force();
        assertEquals("a", promiseKept);
        assertTrue(s.empty());

        try {
            s.pop();
            fail("Popping empty stack should have failed");
        } catch (LxException e) {
            // OK
        } catch (Exception e) {
            fail("Popping empty stack: unexpected exception: " + e);
        }

    }

    private void setPrefixMapping(Stack s, String prefix, String uri)
            throws LxException {
        Stack.Mappings pm = s.newMapping();
        pm.mapPrefix(prefix, uri);
        s.push(pm);
    }
    private void setFeature(Stack s, String uri, boolean value)
            throws LxException {
        Stack.Mappings pm = s.newMapping();
        pm.mapFeature(uri, value);
        s.push(pm);
    }
    
    public void testMappings()
            throws Exception {

        Stack s = new Stack();
        
        s.setPrefixMapping("lx", LxReader.LXNAMESPACE);
        s.setFeature(LxReader.LXNAMESPACE+"/one", true);
        
        s.push(new Stack.Element("one", makePromise("a"))); // 1, element "one"
        Stack.Mappings pm = s.newMapping(); //new Stack.Mappings();
        pm.mapPrefix("a", "http://example.ac.uk/one");
        pm.mapPrefix("b", "http://example.ac.uk/two");
        pm.mapPrefix("", "http://example.ac.uk/three");
        pm.mapPrefix(null, "http://example.ac.uk/four");
        pm.mapFeature("http://example.ac.uk/feat/1", true);
        pm.mapFeature("http://example.ac.uk/feat/1a", true);
        s.push(pm);             // 2, mapping

        assertEquals("http://example.ac.uk/one",
                     s.getPrefixMapping("a").getNamespace());
        assertEquals("http://example.ac.uk/two",
                     s.getPrefixMapping("b").getNamespace());
        Stack.PrefixNamespacePair pnp = s.getPrefixMapping("");
        // following prefix is not a documented feature of the API,
        // but this is the current behaviour
        assertEquals("lxns", pnp.getPrefix());
        assertEquals("http://example.ac.uk/three", pnp.getNamespace());
        pnp = s.getPrefixMapping(null);
        assertEquals("lxns1", pnp.getPrefix());
        assertEquals("http://example.ac.uk/four", pnp.getNamespace());

        assertTrue(s.getFeature(LxReader.LXNAMESPACE+"/one"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1a"));

        s.push(new Stack.Element("two", makePromise("b"))); // 3, element "two"
        pm = s.newMapping(); //new Stack.Mappings();
        pm.mapPrefix("b", "http://example.ac.uk/second");
        pm.mapPrefix("c", "http://example.ac.uk/three");
        s.push(pm);             // 4, mapping

        pnp = s.getPrefixMapping("lx");
        assertNotNull(pnp);
        String uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals(LxReader.LXNAMESPACE, uri);

        pnp = s.getPrefixMapping("a");
        assertNotNull(pnp);
        uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals("http://example.ac.uk/one", uri);
        
        pnp = s.getPrefixMapping("b");
        assertNotNull(pnp);
        uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals("http://example.ac.uk/second", uri);

        pnp = s.getPrefixMapping("c");
        assertNotNull(pnp);
        uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals("http://example.ac.uk/three", uri);
                     

        pm = s.newMapping(); //new Stack.Mappings();
        pm.mapFeature("http://example.ac.uk/feat/1", false);
        pm.mapFeature("http://example.ac.uk/feat/2", true);
        s.push(pm);               // 5, feature mapping

        assertTrue(s.getFeature(LxReader.LXNAMESPACE+"/one"));
        assertFalse(s.getFeature("http://example.ac.uk/feat/1"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1a"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/2"));


        setPrefixMapping(s, "b", "uri:nothing"); // 6, mapping
        assertEquals("http://example.ac.uk/one",
                     s.getPrefixMapping("a").getNamespace());
        assertEquals("uri:nothing",
                     s.getPrefixMapping("b").getNamespace());

        setFeature(s, "http://example.ac.uk/feat/1", true); // 7, feature mapping
        assertTrue(s.getFeature("http://example.ac.uk/feat/1"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1a"));
        assertNull(s.pop());    // 7, feature mapping
        assertNull(s.pop());    // 6, mapping of "b"

        assertNull(s.pop());    // 5, feature mapping

        assertEquals("http://example.ac.uk/second",
                     s.getPrefixMapping("b").getNamespace());
        assertNull(s.pop());    // 4, mapping of "b", "c"

        Stack.Promise p = s.pop(); // 3, element 'two'
        assertNotNull(p);
        p.force();
        assertEquals("b", promiseKept);
        assertEquals("http://example.ac.uk/one",
                     s.getPrefixMapping("a").getNamespace());
        assertEquals("http://example.ac.uk/two",
                     s.getPrefixMapping("b").getNamespace());
        assertTrue(s.getFeature("http://example.ac.uk/feat/1"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1a"));


        pnp = s.getPrefixMapping("a");
        assertNotNull(pnp);
        uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals("http://example.ac.uk/one", uri);

        pnp = s.getPrefixMapping("b");
        assertNotNull(pnp);
        uri = pnp.getNamespace();
        assertNotNull(uri);
        assertEquals("http://example.ac.uk/two", uri);
        
        assertNull(s.getPrefixMapping("c"));

        assertTrue(s.getFeature("http://example.ac.uk/feat/1"));
        assertTrue(s.getFeature("http://example.ac.uk/feat/1a"));
        // Since we have popped the stack frame that had the mapping
        // for feature ".../feat/2", asking for this feature now
        // returns false always (as opposed to SAXNotRecognizedException).
        assertFalse(s.getFeature("http://example.ac.uk/feat/2"));

        s.pop();                // 2, mapping

        assertNull(s.getPrefixMapping("a"));
        assertNull(s.getPrefixMapping("b"));

        assertTrue(s.getFeature(LxReader.LXNAMESPACE+"/one"));
        String[] featurearray = {
            "http://example.ac.uk/feat/2",
            "http://example.ac.uk/feat/1",
            "http://example.ac.uk/feat/1a",
            "http://nothing",
        };
        for (int i=0; i<featurearray.length; i++) {
            // These features are now out of the stack
            assertFalse(s.getFeature(featurearray[i]));
        }
        

        s.pop();                // 1, element "one"
        assertTrue(s.empty());
        assertEquals(LxReader.LXNAMESPACE,
                     s.getPrefixMapping("lx").getNamespace());
        assertNull(s.getPrefixMapping("a"));
        assertNull(s.getPrefixMapping("b"));
        assertNull(s.getPrefixMapping("c"));

        assertTrue(s.getFeature(LxReader.LXNAMESPACE+"/one"));

        // Following tests should fail
        try {
            s.newMapping().mapPrefix("lxtest", "http://x.com/nothing");
            fail("Attempt to map prefix starting lx did not fail");
        } catch (LxException e) {
            // OK
        }
        // But this one should work, since we're mapping it to LXNAMESPACE
        s.newMapping().mapPrefix("lxtest", LxReader.LXNAMESPACE);
    }

    public void testElementMapping()
            throws Exception {

        Stack s = new Stack();
        Stack.Mappings m = s.newMapping();
        QName qea = new QName("ea", s);
        QName qeb = new QName("eb", s);
        QName qec = new QName("ec", s);

        m.mapElementDecl
                (qea,
                 (LxList)getLxObject
                 ("(a1 \"a2\" (a3 #f) (\"a4\" #t) (a5 #t) a6)"));
        s.push(m);

        Stack.AttributeInfo ai = s.getElementDecl(qea);
        assertTrue(ai.allowedAttribute(new QName("a1", s)));
        assertTrue(ai.allowedAttribute(new QName("a2", s)));
        assertTrue(ai.allowedAttribute(new QName("a3", s)));
        assertTrue(ai.allowedAttribute(new QName("a4", s)));
        assertTrue(ai.allowedAttribute(new QName("a5", s)));
        assertTrue(ai.allowedAttribute(new QName("a6", s)));
        assertFalse(ai.allowedAttribute(new QName("x1", s)));

        java.util.Iterator ri = ai.allRequired();
        assertTrue(ri.hasNext());
        Object o = ri.next();
        assertTrue(o instanceof QName);
        assertEquals("a4", ((QName)o).localName());
        assertTrue(ri.hasNext());
        o = ri.next();
        assertTrue(o instanceof QName);
        assertEquals("a5", ((QName)o).localName());
        assertFalse(ri.hasNext());

        s.pop();


        m = s.newMapping();
        m.mapPrefix("a", "urn:1");
        s.push(m);

        // Register the element's allowed attribute by referring to the
        // "urn:1" namespace with the "a" prefix
        Stack.Mappings m2 = s.newMapping();
        QName qna = new QName("a:e1", s);
        m2.mapElementDecl(qna, (LxList)getLxObject("(a:a1 a2)"));
        s.push(m2);

        Stack.Mappings m3 = s.newMapping();
        m.mapPrefix("b", "urn:1");
        s.push(m3);

        // Now refer to the elements and attributes using the new prefix,
        // which maps to the same namespace, and so should be equivalent.
        ai = s.getElementDecl(new QName("b:e1", s));
        assertNotNull(ai);
        assertTrue(ai.allowedAttribute(new QName("b:a1", s)));
        assertTrue(ai.allowedAttribute(new QName("a:a1", s)));
        assertTrue(ai.allowedAttribute(new QName("a2", s)));
        assertFalse(ai.allowedAttribute(new QName("a3", s)));
        assertFalse(ai.allowedAttribute(new QName("b:a2", s)));

    }

    /**
     * Assert that a condition is false.
     * Older versions of JUnit don't have this.
     */
    static public void assertFalse(boolean test) {
        assertTrue(! test);
    }

    private Stack.Promise makePromise(final String s) {
        return new Stack.Promise() {
                protected void doit() {
                    promiseKept = s;
                }
            };
    }

    static private LxObject getLxObject(Tokenizer t) 
            throws LxException {
        return LxObject.readLxObject(t);
    }
    
    static private LxObject getLxObject(String s) 
            throws LxException {
        return getLxObject(new Tokenizer(s));
    }
}
