// This file is part of Lx.
//
// Copyright 2003, 2004, Norman Gray (norman@astro.gla.ac.uk).
// Released under the terms of the GNU General Public Licence:
// see the file LICENCE for details.


package uk.me.nxg.lx.tokenizer;

import junit.framework.Assert;
import junit.framework.TestCase;

import uk.me.nxg.lx.tokenizer.Tokenizer.Token;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintWriter;

public class TokenizerTest
	extends TestCase {

    public TokenizerTest(String name) {
	super(name);
    }

    public void testStringTokenizer() 
	    throws Exception {
	Tokenizer t = new Tokenizer
		("(hello \"there\" plainatt: prefix:att: \"keywordish:\" #t)");

	assertToken(Token.OPENPAREN, t);
	assertToken(Token.SYMBOL, "hello", t);
	assertToken(Token.STRING, "there", t);
	assertToken(Token.KEYWORD, "plainatt", t);
	assertToken(Token.KEYWORD, "prefix:att", t);
	t.ungetToken(Token.KEYWORD);
	assertToken(Token.KEYWORD, "prefix:att", t);
	t.ungetToken(Token.STRING);
	assertToken(Token.STRING, "prefix:att", t);
	assertToken(Token.STRING, "keywordish:", t);
	assertToken(Token.TOKEN, "t", t);
	assertToken(Token.CLOSEPAREN, t);
        assertToken(Token.EOF, t);
	assertEquals(1, t.getLineNumber());
    }

    public void testPushback() 
	    throws Exception {
	Tokenizer t = new Tokenizer
		("(hello plainatt: finally)");

	assertToken(Token.OPENPAREN, t);
	assertToken(Token.SYMBOL, "hello", t);
	assertToken(Token.KEYWORD, "plainatt", t);
        // simple unget
        t.ungetToken(Token.KEYWORD);
	assertToken(Token.KEYWORD, "plainatt", t);
        // push back as a different type
        t.ungetToken(Token.STRING);
        assertToken(Token.STRING, "plainatt", t);
        // push back different type and string
        t.ungetToken(Token.KEYWORD, "different");
        assertToken(Token.KEYWORD, "different", t);
        // push back two tokens, one of them a string
        t.ungetToken(Token.KEYWORD);
        t.ungetToken(Token.OPENPAREN);
        assertToken(Token.OPENPAREN, t);
        assertToken(Token.KEYWORD, "different", t);
        // push back two, both strings
        t.ungetToken(Token.KEYWORD, "one");
        t.ungetToken(Token.STRING, "two");
        assertToken(Token.STRING, "two", t);
        assertToken(Token.KEYWORD, "one", t);
        // and finish off the list -- shouldn't be disturbed by all
        // this interim ungetting 
        assertToken(Token.SYMBOL, "finally", t);
        assertToken(Token.CLOSEPAREN, t);
        assertToken(Token.EOF, t);
	assertEquals(1, t.getLineNumber());
    }

    public void testQuoting()
            throws Exception {
	Tokenizer t = new Tokenizer
                ("\"hello\" \"hell\\\"o\" \"hell[]o\" \"hell[o\" \"hell\\no\" [hello] [[]] [hell[o]] [hell\"o] [hell\\\\o] [hell\\]] [hell\\[] [hell\no]");

        String[] expected = {
            "hello",
            "hell\"o",
            "hell[]o",
            "hell[o",
            "hell\no",
            "hello",
            "[]",
            "hell[o]",
            "hell\"o",
            "hell\\o",
            "hell]",
            "hell[",
            "hell\no",
        };

        for (int i=0; i<expected.length; i++) {
            assertToken(Token.STRING, expected[i], t);
        }

        String[] badstrings = {
            "\"hello",          // EOF
            "\"hello\\",        // no character following escape char -- EOF
            "\"hello\\\"",      // final quote escaped -- EOF again
            "[hell[o]",         // bad nesting
        };
        for (int i=0; i<badstrings.length; i++) {
            try {
                t = new Tokenizer(badstrings[i]);
                Token tok = t.getToken();
                fail("Tokenizer reading string <" + badstrings[i]
                     + "> did not fail as expected");
            } catch (uk.me.nxg.lx.LxException e) {
                // OK
            }
        }
    }

    public void testFileTokenizer()
	    throws Exception {
	// Create a test file
	File testfile = new File(System.getProperty("build.tests"),
				 "test1.lx");
	PrintWriter f = new PrintWriter(new FileOutputStream(testfile));
	f.println("(hello there:");
	f.println("   \"how \\\" are\" ;; this line isn't seen");
	f.println(";");
	f.println("   you \"?\"");
	f.println();
	f.println(" #now");
	f.close();

	// Use it
	Tokenizer t = new Tokenizer(testfile);

	assertToken(Token.OPENPAREN, t);
	assertEquals(1, t.getLineNumber());
	assertToken(Token.SYMBOL, "hello", t);
	assertEquals(1, t.getLineNumber());
	assertToken(Token.KEYWORD, "there", t);
	assertEquals(1, t.getLineNumber());
	assertToken(Token.STRING, "how \" are", t);
	assertEquals(2, t.getLineNumber());
	assertToken(Token.SYMBOL, "you", t);
	assertEquals(4, t.getLineNumber());
	assertToken(Token.STRING, "?", t);
	assertEquals(4, t.getLineNumber());
	assertToken(Token.TOKEN, "now", t);
	assertEquals(6, t.getLineNumber());

	// Delete it
	testfile.delete();
    }

    public void testTorture()
            throws Exception {
        Tokenizer t = new Tokenizer
                ("s s\"q\" s[q][s=\"cmd 2>]&[1|grep '\\\\(^File\\\\|Examining\\\\)'|grep 'ok'\"]");
        assertToken(Token.SYMBOL, "s", t);
        assertToken(Token.SYMBOL, "s", t);
        assertToken(Token.STRING, "q", t);
        assertToken(Token.SYMBOL, "s", t);
        assertToken(Token.STRING, "q", t);
        assertToken(Token.STRING, "s=\"cmd 2>", t);
        assertToken(Token.SYMBOL, "&", t);
        assertToken(Token.STRING, "1|grep '\\(^File\\|Examining\\)'|grep 'ok'\"", t);
    }

    private void assertToken(Tokenizer.Token expectedToken, Tokenizer t)
            throws Exception {
        assertToken(expectedToken, null, t);
    }

    private void assertToken(Tokenizer.Token expectedToken, String expectedValue,
			     Tokenizer t) 
	    throws Exception {
	assertEquals(expectedToken, t.getToken());
	if (expectedValue != null)
            assertEquals(expectedValue, t.value());
    }
}
